#coding:utf-8
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/protocol/fbsp.py
# DESCRIPTION:    Firebird Butler Service Protocol
#                 See https://firebird-butler.readthedocs.io/en/latest/rfc/4/FBSP.html
# CREATED:        21.2.2019
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2019 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________.

"""Saturnin reference implementation of Firebird Butler Service Protocol.

See https://firebird-butler.readthedocs.io/en/latest/rfc/4/FBSP.html
"""

from __future__ import annotations
from typing import Type, Any, List, Dict, Callable
import uuid
from struct import pack, unpack
from enum import IntEnum, IntFlag
from traceback import format_exception
import zmq
from firebird.base.signal import eventsocket
from firebird.base.protobuf import ProtoMessage, create_message
from saturnin.base import Token, RoutingID, State, ButlerInterface,  Channel, Protocol, \
     Message, Session, ANY, INVALID,  ServiceDescriptor, AgentDescriptor, PeerDescriptor, \
     StopError, ServiceError, InvalidMessageError, TZMQMessage

# Message header
#: FBSP protocol control frame :mod:`struct` format
HEADER_FMT_FULL = '!4sBBH8s'
#: FBSP protocol control frame :mod:`struct` format without FOURCC
HEADER_FMT = '!4xBBH8s'
#: FBSP protocol identification (FOURCC)
FOURCC = b'FBSP'
#: FBSP protocol version mask
VERSION_MASK = 7
#: FBSP protocol error mask
ERROR_TYPE_MASK = 31

# Protobuf messages

PROTO_HELLO = 'firebird.butler.FBSPHelloDataframe'
PROTO_WELCOME = 'firebird.butler.FBSPWelcomeDataframe'
PROTO_CANCEL_REQ = 'firebird.butler.FBSPCancelRequests'
PROTO_STATE_INFO = 'firebird.butler.FBSPStateInformation'
PROTO_ERROR = 'firebird.butler.ErrorDescription'

# Enums

class MsgType(IntEnum):
    """FBSP Message Type"""
    UNKNOWN = 0 # Not a valid option, defined only to handle undefined values
    HELLO = 1   # initial message from client
    WELCOME = 2 # initial message from service
    NOOP = 3    # no operation, used for keep-alive & ping purposes
    REQUEST = 4 # client request
    REPLY = 5   # service response to client request
    DATA = 6    # separate data sent by either client or service
    CANCEL = 7  # cancel request
    STATE = 8   # operating state information
    CLOSE = 9   # sent by peer that is going to close the connection
    ERROR = 31  # error reported by service

class MsgFlag(IntFlag):
    """FBSP message flag"""
    NONE = 0
    ACK_REQ = 1
    ACK_REPLY = 2
    MORE = 4

class ErrorCode(IntEnum):
    """FBSP Error Code"""
    # Errors indicating that particular request cannot be satisfied
    INVALID_MESSAGE = 1
    PROTOCOL_VIOLATION = 2
    BAD_REQUEST = 3
    NOT_IMPLEMENTED = 4
    ERROR = 5
    INTERNAL_ERROR = 6
    REQUEST_TIMEOUT = 7
    TOO_MANY_REQUESTS = 8
    FAILED_DEPENDENCY = 9
    FORBIDDEN = 10
    UNAUTHORIZED = 11
    NOT_FOUND = 12
    GONE = 13
    CONFLICT = 14
    PAYLOAD_TOO_LARGE = 15
    INSUFFICIENT_STORAGE = 16
    REQUEST_CANCELLED = 17
    # Fatal errors indicating that connection would/should be terminated
    SERVICE_UNAVAILABLE = 2000
    FBSP_VERSION_NOT_SUPPORTED = 2001

def bb2h(value_hi: int, value_lo: int) -> int:
    """Compose two bytes into word value.
    """
    return unpack('!H', pack('!BB', value_hi, value_lo))[0]

def msg_bytes(msg: Union[bytes, bytearray, zmq.Frame]) -> Union[bytes, bytearray]:
    """Return message frame as bytes.
    """
    return msg.bytes if isinstance(msg, zmq.Frame) else msg

class FBSPMessage(Message):
    """Firebird Butler Service Protocol (FBSP) Message.
    """
    def __init__(self):
        #: Type of message
        self.msg_type: MsgType = MsgType.UNKNOWN
        #: Message flags
        self.flags: MsgFlag = MsgFlag.NONE
        #: Data associated with message
        self.type_data: int = 0
        #: Message token
        self.token: Token = bytearray(8)
    def __str__(self):
        return f"{self.__class__.__qualname__}[{self.msg_type.name}]"
    __repr__ = __str__
    def _unpack_data(self, data: List) -> None:
        """Called when all fields of the message are set. Usefull for data deserialization.
        """
    def _pack_data(self) -> List:
        """Called when serialization is requested.
        """
        return []
    def _set_hdr(self, header: bytes) -> None:
        """Initialize new message from header.
        """
        control_byte, flags, self.type_data, self.token = unpack(HEADER_FMT, header)
        self.msg_type = MsgType(control_byte >> 3)
        self.flags = MsgFlag(flags)
    def from_zmsg(self, zmsg: TZMQMessage) -> None:
        """Populate message data from sequence of ZMQ data frames.

        Arguments:
            zmsg: Sequence of frames that should be deserialized.

        Raises:
            InvalidMessageError: If message is not a valid protocol message.
        """
        try:
            zmsg.pop(0) # header is already set by message_factory
            #control_byte, flags, self.type_data, self.token = unpack(HEADER_FMT, zmsg.pop(0))
            #self.msg_type = MsgType(control_byte >> 3)
            #self.flags = MsgFlag(flags)
            if MsgFlag.ACK_REPLY not in self.flags:
                self._unpack_data(zmsg)
        except Exception as exc:
            raise InvalidMessageError("Invalid message") from exc
    def as_zmsg(self) -> TZMQMessage:
        """Returns message as sequence of ZMQ data frames.
        """
        zmsg = [self.get_header()]
        if MsgFlag.ACK_REPLY not in self.flags:
            zmsg.extend(self._pack_data())
        return zmsg
    def clear(self) -> None:
        """Clears message data.

        Important:
            The `msg_type` remains the same (i.e. you can't change the message type once
            message instance was created), but all other data are cleared.
        """
        self.type_data = 0
        self.flags = MsgFlag(0)
        self.token: Token = bytearray(8)
    def copy(self) -> Message:
        """Returns copy of the message.
        """
        msg: FBSPMessage = self.__class__()
        msg.from_zmsg(self.as_zmsg())
        return msg
    def get_keys(self) -> Iterable:
        """Returns iterable of dictionary keys to be used with `Protocol.handlers`.
        Keys must be provided in order of precedence (from more specific to general).
        """
        return [self.msg_type, ANY]
    def get_header(self) -> bytes:
        """Return message header (FBSP control frame).
        """
        return pack(HEADER_FMT_FULL, FOURCC, (self.msg_type << 3) | 1, self.flags,
                    self.type_data, self.token)
    def has_more(self) -> bool:
        """Returns True if message has `MORE` flag set.
        """
        return MsgFlag.MORE in self.flags
    def has_ack_req(self) -> bool:
        """Returns True if message has ACK_REQ flag set.
        """
        return MsgFlag.ACK_REQ in self.flags
    def has_ack_reply(self) -> bool:
        """Returns True if message has ASK_REPLY flag set.
        """
        return MsgFlag.ACK_REPLY in self.flags
    def set_flag(self, flag: MsgFlag) -> None:
        """Set flag specified by `flag` mask.
        """
        self.flags |= flag
    def clear_flag(self, flag: MsgFlag) -> None:
        """Clear flag specified by `flag` mask.
        """
        self.flags &= ~flag

class HandshakeMessage(FBSPMessage):
    """Base FBSP client/service handshake message (`HELLO` or `WELCOME`). The message
    includes basic information about the Peer.
    """
    def __init__(self):
        super().__init__()
        #: Parsed protobuf message with peer information
        self.data: ProtoMessage = None
    def _unpack_data(self, data: List) -> None:
        """Called when all fields of the message are set. Usefull for data deserialization.
        """
        self.data.ParseFromString(data.pop(0))
    def _pack_data(self) -> List:
        """Called when serialization is requested.
        """
        return [self.data.SerializeToString()]
    def clear(self) -> None:
        """Clears message attributes.
        """
        super().clear()
        self.data.Clear()

class HelloMessage(HandshakeMessage):
    """The `HELLO` message is a Client request to open a Connection to the Service.
    The message includes basic information about the Client and Connection parameters
    required by the Client.
    """
    def __init__(self):
        super().__init__()
        #:`FBSPHelloDataframe` protobuf message with peer information
        self.data = create_message(PROTO_HELLO)

class WelcomeMessage(HandshakeMessage):
    """The `WELCOME` message is the response of the Service to the `HELLO` message sent by
    the Client, which confirms the successful creation of the required Connection and announces
    basic parameters of the Service and the Connection.
    """
    def __init__(self):
        super().__init__()
        #: `FBSPWelcomeDataframe` protobuf message with peer information
        self.data = create_message(PROTO_WELCOME)

class APIMessage(FBSPMessage):
    """Base FBSP client/service API message (`REQUEST`, `REPLY`, `STATE`).
    The message includes information about the API call (interface ID and API Code).
    """
    def __init__(self):
        super().__init__()
        self.data: List = []
    def _unpack_data(self, data: List) -> None:
        """Called when all fields of the message are set. Usefull for data deserialization.
        """
        self.data.extend(i.bytes if isinstance(i, zmq.Frame) else i for i in data)
    def _pack_data(self) -> List:
        """Called when serialization is requested.
        """
        return self.data
    def clear(self) -> None:
        """Clears message attributes.
        """
        super().clear()
        self.__interface_id = 0
        self.__api_code = 0
        self.data.clear()
    @property
    def interface_id(self) -> int:
        "Interface ID (high byte of Request Code)"
        return unpack('!BB', pack('!H', self.type_data))[0]
    @interface_id.setter
    def interface_id(self, value: int) -> None:
        self.type_data = bb2h(value, self.api_code)
    @property
    def api_code(self) -> int:
        "API Code (lower byte of Request Code)"
        return unpack('!BB', pack('!H', self.type_data))[1]
    @api_code.setter
    def api_code(self, value: int) -> None:
        self.type_data = bb2h(self.interface_id, value)
    @property
    def request_code(self) -> int:
        "Request Code (Interface ID + API Code)"
        return self.type_data

class StateMessage(APIMessage):
    """The `STATE` message is a Client request to the Service.
    """
    def __init__(self):
        super().__init__()
        #: `FBSPStateInformation` protobuf message with state information
        self.state_info: ProtoMessage = create_message(PROTO_STATE_INFO)
    def _unpack_data(self, data: List) -> None:
        """Called when all fields of the message are set. Usefull for data deserialization.
        """
        self.state_info.ParseFromString(data.pop(0))
        super()._unpack_data(data)
    def _pack_data(self) -> List:
        """Called when serialization is requested.
        """
        return [self.state_info.SerializeToString()]
    def clear(self) -> None:
        """Clears message attributes.
        """
        super().clear()
        self.state_info.Clear()
    @property
    def state(self) -> State:
        "Service state"
        return State(self.state_info.state)
    @state.setter
    def state(self, value: State) -> None:
        self.state_info.state = value.value

class DataMessage(FBSPMessage):
    """The `DATA` message is intended for delivery of arbitrary data between connected peers."""
    def __init__(self):
        super().__init__()
        #: Data payload
        self.data: List[bytes] = []
    def _unpack_data(self, data: List) -> None:
        """Called when all fields of the message are set. Usefull for data deserialization.
        """
        self.data.extend(i.bytes if isinstance(i, zmq.Frame) else i for i in data)
    def _pack_data(self) -> List:
        """Called when serialization is requested.
        """
        return self.data
    def clear(self) -> None:
        """Clears message attributes.
        """
        super().clear()
        self.data.clear()

class CancelMessage(FBSPMessage):
    """The `CANCEL` message represents a request for a Service to stop processing the previous
    request from the Client.
    """
    def __init__(self):
        super().__init__()
        #: `.FBSPCancelRequests` protobuf message
        self.reqest: ProtoMessage = create_message(PROTO_CANCEL_REQ)
    def _unpack_data(self, data: List) -> None:
        """Called when all fields of the message are set. Usefull for data deserialization.
        """
        self.reqest.ParseFromString(data.pop(0))
    def _pack_data(self) -> List:
        """Called when serialization is requested.
        """
        return [self.reqest.SerializeToString()]
    def clear(self) -> None:
        """Clears message attributes.
        """
        super().clear()
        self.reqest.Clear()

class ErrorMessage(FBSPMessage):
    """The `ERROR` message notifies the Client about error condition detected by Service.
    """
    def __init__(self):
        super().__init__()
        #: List of `.ErrorDescription` protobuf messages with error information
        self.errors: List[ProtoMessage] = []
    def _unpack_data(self, data: List) -> None:
        """Called when all fields of the message are set. Usefull for data deserialization.
        """
        while data:
            err = create_message(PROTO_ERROR)
            err.ParseFromString(msg_bytes(data.pop(0)))
            self.errors.append(err)
    def _pack_data(self) -> List:
        """Called when serialization is requested.
        """
        return [err.SerializeToString() for err in self.errors]
    def clear(self) -> None:
        """Clears message attributes.
        """
        super().clear()
        self.errors.clear()
    def add_error(self) -> ProtoMessage:
        """Return newly created `ErrorDescription` associated with message."""
        err = create_message(PROTO_ERROR)
        self.errors.append(err)
        return err
    def note_exception(self, exc: Exception) -> None:
        """Store information from exception into `.ErrorMessage`.
        """
        to_note = exc
        while to_note:
            errdesc = self.add_error()
            if hasattr(to_note, 'code'):
                errdesc.code = getattr(to_note, 'code')
            errdesc.description = str(to_note)
            if not isinstance(to_note, StopError):
                errdesc.annotation['traceback'] = \
                    ''.join(format_exception(to_note.__class__, to_note,
                                             to_note.__traceback__, chain=False))
            to_note = to_note.__cause__
    @property
    def error_code(self) -> ErrorCode:
        "Error code"
        return ErrorCode(self.type_data >> 5)
    @error_code.setter
    def error_code(self, value: ErrorCode) -> None:
        self.type_data = (value.value << 5) | (self.type_data & ERROR_TYPE_MASK)
    @property
    def relates_to(self) -> MsgType:
        "Message type this error relates to"
        return MsgType(self.type_data & ERROR_TYPE_MASK)
    @relates_to.setter
    def relates_to(self, value: MsgType) -> None:
        self.type_data &= ~ERROR_TYPE_MASK
        self.type_data |= value.value

class _FBSP(Protocol):
    """4/FBSP - Firebird Butler Service Protocol
    """
    #: string with protocol OID (dot notation).
    OID: str =  '1.3.6.1.4.1.53446.1.5.1'
    # iso.org.dod.internet.private.enterprise.firebird.butler.protocol.fbsp
    #: UUID instance that identifies the protocol.
    UID: uuid.UUID = uuid.uuid5(uuid.NAMESPACE_OID, OID)
    #: Mapping from message type to specific Message class
    MESSAGE_MAP: Dict[MsgType, FBSPMessage] = \
        {MsgType.HELLO: HelloMessage,
         MsgType.WELCOME: WelcomeMessage,
         MsgType.NOOP: FBSPMessage,
         MsgType.REQUEST: APIMessage,
         MsgType.REPLY: APIMessage,
         MsgType.DATA: DataMessage,
         MsgType.CANCEL: CancelMessage,
         MsgType.STATE: StateMessage,
         MsgType.CLOSE: FBSPMessage,
         MsgType.ERROR: ErrorMessage,
         }
    def __init__(self, *, session_type: Type[Session]):
        """
        Arguments:
            session_type: Class for session objects.
            batch_size: Default batch size.
        """
        super().__init__(session_type=session_type)
        self.message_factory = self.__message_factory
    def __message_factory(self, zmsg: TZMQMessage=None) -> Message:
        """Internal message factory.
        """
        assert zmsg is not None and len(zmsg) >= 1, "ZMQ message header is required to create correct message instance"
        msg = self.MESSAGE_MAP[MsgType(int.from_bytes(zmsg[0][4:5], 'big') >> 3)]()
        msg._set_hdr(zmsg[0])
        return msg
    def validate(self, zmsg: TZMQMessage) -> None:
        """Verifies that sequence of ZMQ data frames is a valid protocol message.

        If this validation passes without exception, then `.convert_msg()` of the same
        message must be successful as well.

        Important:
            Implementation in base Protocol performs no validation and always returns True.

        Arguments:
            zmsg:   ZeroMQ multipart message.

        Raises:
            InvalidMessageError: If ZMQ message is not a valid protocol message.
        """
        if not zmsg:
            raise InvalidMessageError("Empty message")
        fbsp_header = zmsg[0]
        if len(fbsp_header) != 16:
            raise InvalidMessageError("Message header must be 16 bytes long")
        try:
            fourcc, control_byte, flags, type_data, token = unpack(HEADER_FMT_FULL, fbsp_header)
        except Exception as exp:
            raise InvalidMessageError("Can't parse the control frame") from exp
        if fourcc != FOURCC:
            raise InvalidMessageError("Invalid FourCC")
        if (control_byte & VERSION_MASK) != self.REVISION:
            raise InvalidMessageError("Invalid protocol version")
        if (flags | 7) > 7:
            raise InvalidMessageError("Invalid flags")
        msg_type = control_byte >> 3
        try:
            if msg_type == 0:
                raise ValueError()
            message_type = MsgType(msg_type)
        except ValueError:
            raise InvalidMessageError(f"Illegal message type {msg_type}")
        #
        if message_type in (MsgType.REQUEST, MsgType.REPLY, MsgType.STATE):
            # Check request_code validity
            pass
        if message_type is MsgType.ERROR:
            try:
                ErrorCode(type_data >> 5)
            except ValueError:
                raise InvalidMessageError(f"Unknown ERROR code: {type_data >> 5}")
            if MsgType(type_data & ERROR_TYPE_MASK) not in (MsgType.UNKNOWN, MsgType.HELLO,
                                                            MsgType.REQUEST, MsgType.DATA,
                                                            MsgType.CANCEL):
                raise InvalidMessageError("Invalid request code in ERROR message")
            if len(zmsg) > 1:
                frame = create_message(PROTO_ERROR)
                for i, segment in enumerate(zmsg[1:]):
                    try:
                        frame.ParseFromString(msg_bytes(segment))
                        frame.Clear()
                    except Exception as exc:
                        raise InvalidMessageError(f"Invalid ERROR message data frame: {i}") from exc
        elif message_type is MsgType.HELLO:
            try:
                create_message(PROTO_HELLO).ParseFromString(msg_bytes(zmsg[1]))
            except Exception as exc:
                raise InvalidMessageError("Invalid HELLO message data frame") from exc
        elif message_type is MsgType.WELCOME:
            try:
                create_message(PROTO_WELCOME).ParseFromString(msg_bytes(zmsg[1]))
            except Exception as exc:
                raise InvalidMessageError("Invalid WELCOME message data frame") from exc
        elif message_type is MsgType.NOOP:
            if len(zmsg) > 1:
                raise InvalidMessageError("Data frames not allowed for NOOP message")
        elif message_type is MsgType.CANCEL:
            if len(zmsg) > 2:
                raise InvalidMessageError("CANCEL message must have exactly one data frame")
            try:
                create_message(PROTO_CANCEL_REQ).ParseFromString(msg_bytes(zmsg[2]))
            except Exception as exc:
                raise InvalidMessageError("Invalid CANCEL message data frame") from exc
        elif message_type is MsgType.STATE:
            if len(zmsg) > 2:
                raise InvalidMessageError("STATE message must have exactly one data frame")
            try:
                create_message(PROTO_STATE_INFO).ParseFromString(msg_bytes(zmsg[2]))
            except Exception as exc:
                raise InvalidMessageError("Invalid STATE message data frame") from exc
    def create_message_for(self, message_type: MsgType, token: Token=bytes(8),
                           type_data: int=0, flags: MsgFlag=MsgFlag.NONE) -> FBSPMessage:
        """Create new `.FBDPMessage` child class instance for particular FBSP message type.

        Arguments:
            message_type: Type of message to be created
            token:        Message token
            type_data:    Message control data
            flags:        Flags

        Returns:
            New `.FBDPMessage` instance.
        """
        return self.message_factory([pack(HEADER_FMT_FULL, FOURCC, (message_type << 3) | 1,
                                          flags, type_data, token)])
    def create_ack_reply(self, msg: FBSPMessage) -> FBSPMessage:
        """Returns new message that is an ACK-REPLY response message.

        Arguments:
            msg: Message to be acknowledged.
        """
        return self.message_factory([pack(HEADER_FMT_FULL, FOURCC,
                                          (msg.msg_type << 3) | 1,
                                          (msg.flags & ~MsgFlag.ACK_REQ) | MsgFlag.ACK_REPLY,
                                          msg.type_data, msg.token)])
    def create_data_for(self, msg: APIMessage) -> DataMessage:
        """Create new DataMessage for reply to specific reuest message.

        Arguments:
            message: Request message instance that data relates to
        """
        return self.create_message_for(MsgType.DATA, msg.token)

class FBSPSession(Session):
    """FBSP session that holds information about attached peer.
    """
    def __init__(self):
        super().__init__()
        #: `.HelloMessage` or `.WelcomeMessage` received from peer.
        self.greeting: HelloMessage = None
        #: Client peer ID for service, Service agent ID for client
        self.partner_uid: uuid.UUID = None

class FBSPService(_FBSP):
    """4/FBSP - Firebird Butler Service Protocol - Service side.
    """
    def __init__(self, *, session_type: Type[FBSPSession]=FBSPSession,
                 service: ServiceDescriptor, peer: PeerDescriptor):
        """
        Arguments:
            session_type: Class for session objects.
            agent: Agent descriptor for service.
            peer: Peer descriptor for service.
            interfaces: List of service interface enumerations.
        """
        super().__init__(session_type=session_type)
        self.handlers.update({MsgType.HELLO: self.handle_hello_msg,
                              MsgType.REQUEST: self.handle_request_msg,
                              MsgType.CANCEL: self.handle_cancel_msg,
                              MsgType.NOOP: self.handle_noop_msg,
                              MsgType.DATA: self.handle_data_msg,
                              MsgType.CLOSE: self.handle_close_msg,
                              MsgType.REPLY: self.handle_ack_reply,
                              MsgType.STATE: self.handle_ack_reply,
                              MsgType.WELCOME: self.handle_invalid_msg,
                              })
        self.api_handlers: Dict[Any, Callable] = {}
        self._apis: List[ButlerInterface] = []
        self._apis.extend(service.api)
        self.welcome_df = create_message(PROTO_WELCOME)
        self.welcome_df.instance.uid = peer.uid.bytes
        self.welcome_df.instance.pid = peer.pid
        self.welcome_df.instance.host = peer.host
        self.welcome_df.service.uid = service.agent.uid.bytes
        self.welcome_df.service.name = service.agent.name
        self.welcome_df.service.version = service.agent.version
        self.welcome_df.service.classification = service.agent.classification
        self.welcome_df.service.vendor.uid = service.agent.vendor_uid.bytes
        self.welcome_df.service.platform.uid = service.agent.platform_uid.bytes
        self.welcome_df.service.platform.version = service.agent.platform_version
        for i in range(len(self._apis)):
            intf = self.welcome_df.api.add()
            intf.number = i
            intf.uid = self._apis[i].get_uid().bytes
    def accept_new_session(self, channel: Channel, routing_id: RoutingID, msg: FBSPMessage) -> bool:
        """Validates incoming message that initiated new session/transmission.

        Calls `on_accept_client` event handler that may reject the client by raising `StopError`
        exception with appropriate error code to be sent to the client in ERROR message.

        Arguments:
            channel:    Channel that received the message.
            routing_id: Routing ID of the sender.
            msg:        Received message.
        """
        err_code = None
        err = None
        try:
            if msg.msg_type is not MsgType.HELLO:
                raise StopError("Expecting HELLO message", code=ErrorCode.PROTOCOL_VIOLATION)
            self.on_accept_client(channel, msg)
        except StopError as exc:
            err_code = getattr(exc, 'code', ErrorCode.ERROR)
        except Exception as exc:
            err_code = getattr(exc, 'code', ErrorCode.INTERNAL_ERROR)
            err = exc
        if err_code is not None:
            session = Session()
            session.routing_id = routing_id
            self.send_error(channel, session, msg, err_code, err)
            return False
        return True
    def register_api_handler(self, api_code: ButlerInterface, handler: Callable) -> None:
        """Register handler for REQUEST message for particular service API code.
        """
        _hndcheck.service_handler = handler
        _hndcheck.service_handler = None
        self.api_handlers[bb2h(self._apis.index(api_code.__class__), api_code.value)] = handler
    def create_welcome_reply(self, msg: HelloMessage) -> WelcomeMessage:
        """Create new WelcomeMessage that is a reply to client's HELLO.

        Arguments:
            hello:  `.HelloMessage` from the client
        """
        return self.create_message_for(MsgType.WELCOME, msg.token)
    def create_error_for(self, msg: FBSPMessage, error_code: ErrorCode) -> ErrorMessage:
        """Create new ErrorMessage that relates to specific message.

        Arguments:
            message:    `.Message` instance that error relates to
            error_code: Error code
        """
        err = self.create_message_for(MsgType.ERROR, msg.token)
        err.relates_to = msg.msg_type
        err.error_code = error_code
        return err
    def create_reply_for(self, msg: APIMessage) -> APIMessage:
        """Create new ReplyMessage for specific RequestMessage.

        Arguments:
            message: Request message instance that reply relates to
            value:   State code
        """
        return self.create_message_for(MsgType.REPLY, msg.token, msg.type_data)
    def create_state_for(self, msg: APIMessage, state: State) -> StateMessage:
        """Create new StateMessage that relates to specific RequestMessage.

        Arguments:
            message: Request message instance that state relates to
            state:   State code
        """
        msg = self.create_message_for(MsgType.STATE, msg.token, msg.type_data)
        msg.state = state
        return msg
    def handle_exception(self, channel: Channel, session: Session, msg: Message, exc: Exception) -> None:
        """Called by `.handle_msg()` on exception in message handler.

        Sends ERROR message and then calls `on_exception` handler.

        Important:
            The error code is extracted ONLY from `StopError` exception. Other exception
            types will result in `6 - Internal Service Error` error code.
        """
        error_code = getattr(exc, 'code', ErrorCode.ERROR) if isinstance(exc, StopError) \
            else ErrorCode.INTERNAL_ERROR
        self.send_error(channel, session, msg, error_code, exc)
        super().handle_exception(channel, session, msg, exc)
    def handle_invalid_msg(self, channel: Channel, session: FBSPSession, msg: HelloMessage) -> None:
        """Process invalid message received from client.

        Sends ERROR message to the client with error code `2 - Protocol violation`.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.
        """
        raise StopError("Unexpected message", code=ErrorCode.PROTOCOL_VIOLATION)
    def handle_hello_msg(self, channel: Channel, session: FBSPSession, msg: HelloMessage) -> None:
        """Process HELLO message received from client.

        Sends WELCOME message to the client.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.

        Note:
            The HELLO message was preprocessed and approved by `.accept_new_session()`.
        """
        session.greeting = msg
        session.partner_uid = uuid.UUID(bytes=msg.data.instance.uid)
        welcome = self.create_welcome_reply(msg)
        welcome.data.CopyFrom(self.welcome_df)
        channel.send(welcome, session)
    def handle_request_msg(self, channel: Channel, session: FBSPSession, msg: APIMessage) -> None:
        """Process REQUEST message received from client.

        Calls the appropriate API handler set via `.register_api_handler()`.

        Important:
            The event handler must send the ACK-REPLY message if requested. According to
            FBSP specification, the request must be acknowledged at the time the Service
            has positively decided to accept the client’s request and before commencing
            the fulfillment of the request. It may also send the ERROR instead the ACK-REPLY.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.

        Note:
            All exceptions are handled by `handle_exception`.
        """
        handler = self.api_handlers.get(msg.type_data)
        if handler is None:
            raise StopError("", code=ErrorCode.NOT_IMPLEMENTED)
        handler(channel, session, msg, self)
    def handle_cancel_msg(self, channel: Channel, session: FBSPSession, msg: CancelMessage) -> None:
        """Process CANCEL message received from peer.

        Calls `on_cancel` event handler. According to FBSP, the service must send the ERROR
        message with appropriate error code (if request was successfuly cancelled, the code
        must be `17 - Request Cancelled`). To follow the common pattern, the event handler
        must raise an `.StopError` exception with `code` attribute set to the error code
        to be returned in ERROR message.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.

        Raises:
            StopError: With error code `4 - Not Implemented` if event handler is not defined.

        Note:
            All exceptions are handled by `handle_exception`.
        """
        self.on_cancel(channel, session, msg)
        if not self.on_cancel.is_set():
            raise StopError("Request cancellation not supported", code=ErrorCode.NOT_IMPLEMENTED)
    def handle_noop_msg(self, channel: Channel, session: FBSPSession, msg: FBSPMessage) -> None:
        """Process NOOP message received from client.

        If message is an ACK-REPLY, it calls `on_ack_received` event handler.
        Otherwise it sends ACK-REPLY if requested and then calls `on_noop` event handler.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.

        Note:
            All exceptions are handled by `handle_exception`.
        """
        if msg.has_ack_reply():
            self.on_ack_received(channel, session, msg)
            return
        if msg.has_ack_req():
            channel.send(self.create_ack_reply(msg), session)
        self.on_noop(channel, session)
    def handle_data_msg(self, channel: Channel, session: FBSPSession, msg: DataMessage) -> None:
        """Process DATA message received from peer.

        If message is an ACK-REPLY, it calls `on_ack_received` event handler. Otherwise it
        calls `on_data` event handler (that must also handle the ACK-REQ if set).

        If `on_data` event handler is not defined, the handler sends ERROR message
        with code `2 - Protocol violation` by raising the `StopError`.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.

        Note:
            All exceptions are handled by `handle_exception`.
        """
        if msg.has_ack_reply():
            self.on_ack_received(channel, session, msg)
            return
        self.on_data(channel, session, msg)
        if not self.on_data.is_set():
            raise StopError("Unexpected DATA message", code=ErrorCode.PROTOCOL_VIOLATION)
    def handle_close_msg(self, channel: Channel, session: FBSPSession, msg: FBSPMessage) -> None:
        """Process CLOSE message received from client.

        Calls `on_session_closed` event handler and then discards the session.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.

        Important:
            All exceptions raised by event handler are silently ignored.
        """
        try:
            self.on_session_closed(channel, session, msg)
        except:
            # We don't want to handle this via `handle_exception` and we're closing
            # the session anyway
            pass
        finally:
            channel.discard_session(session)
    def handle_ack_reply(self, channel: Channel, session: FBSPSession, msg: FBSPMessage) -> None:
        """Process ACK-REPLY STATE and REPLY messages received from client.

        If message has ACK-REPLY flag set, it calls `on_ack_received` event handler.
        Otherwise it raises StopError with PROTOCOL_VIOLATION error code.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received REPLY or STATE message.

        Note:
            All exceptions are handled by `handle_exception`.
        """
        if msg.has_ack_reply():
            self.on_ack_received(channel, session, msg)
        else:
            raise StopError(f"Client can send {msg.msg_type.name} messages only as ACK-REPLY",
                            code=ErrorCode.PROTOCOL_VIOLATION)
    def send_error(self, channel: Channel, session: FBSPSession, relates_to: FBSPMessage,
                   error_code: ErrorCode, exc: Exception=None) -> None:
        """Sends `ERROR` message to the client associated with session.

        Arguments:
            channel: Channel associate with data pipe.
            session: Session associated with transmission.
            relates_to: FBSP message to which this ERROR is related.
            error_code: Error code.
            exc: Exception that caused the error.
        """
        msg: ErrorMessage = self.create_message_for(MsgType.ERROR, relates_to.token)
        msg.error_code = error_code
        msg.relates_to = relates_to.msg_type
        if exc is not None:
            msg.note_exception(exc)
        channel.send(msg, session)
    def send_close(self, channel: Channel, session: FBSPSession) -> None:
        """Sends `CLOSE` message to the client associated with session and calls
        `on_session_closed` event handler. The message passed to the event handler is
        the initial HELLO message from client.

        Arguments:
            channel: Channel associate with data pipe.
            session: Session associated with transmission.

        Important:
            All exceptions raised by event handler are silently ignored.
        """
        channel.send(self.create_message_for(MsgType.CLOSE, session.greeting.token), session)
        try:
            self.on_session_closed(channel, session, session.greeting)
        except:
            # We don't want to handle this via `handle_exception` and we're closing
            # the session anyway
            pass
    def close(self, channel: Channel):
        """Close all connections to attached clients.
        """
        while channel.sessions:
            _, session = channel.sessions.popitem()
            try:
                self.send_close(channel, session)
            except:
                # channel could be already closed from other side, as we are closing it too
                # we can ignore any send errors
                pass
    @eventsocket
    def on_accept_client(self, channel: Channel, msg: HelloMessage) -> None:
        """Called when HELLO message is received from client.

        Arguments:
            channel: Channel that received the message.
            msg:     Received message.

        The event handler may reject the client by raising the `StopError` exception
        with `code` attribute containing the `ErrorCode` to be returned in ERROR message.

        If event handler does not raise an exception, the client is accepted, new session
        is created and WELCOME message is sent to the client.
        """
    @eventsocket
    def on_cancel(self, channel: Channel, session: FBSPSession, msg: CancelMessage) -> None:
        """Called when CANCEL message is received from client.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.

        The event handler must raise an `.Error` exception with `code` attribute set to
        the error code to be returned in ERROR message. If request was successfuly cancelled,
        the code must be `ErrorCode.REQUEST_CANCELLED`.
        """
    @eventsocket
    def on_noop(self, channel: Channel, session: FBSPSession) -> None:
        """Called when NOOP message is received, and after ACK-REPLY (if requested) is send.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with peer.
        """
    @eventsocket
    def on_data(self, channel: Channel, session: FBSPSession, msg: DataMessage) -> None:
        """Called when DATA message is received.

        Important:
            If handler is defined, it must send the ACK-REPLY if received message has
            ACK-REQ set. It's up to service interface whether ACK-REPLY would be sent
            before, or after data contained within message are processed.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.

        Note:
            All exceptions are handled by `handle_exception`.
        """
    @eventsocket
    def on_session_closed(self, channel: Channel, session: FBSPSession, msg: FBSPMessage,
                          exc: Exception=None) -> None:
        """Called when CLOSE message is received or sent, to release any resources
        associated with current transmission.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with peer.
            msg: Received/sent CLOSE message.
            exc: Exception that caused the error.
        """
    @eventsocket
    def on_ack_received(self, channel: Channel, session: FBSPSession, msg: DataMessage) -> None:
        """Called when ACK-REPLY NOOP, DATA, REPLY or STATE message is received.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received ACK-REPLY message.

        Note:
            All exceptions are handled by `handle_exception`.
        """

class FBSPClient(_FBSP):
    """4/FBSP - Firebird Butler Service Protocol - Client side.

    This is the RAW version of client side FBSP protocol for clients that directly process
    FBSP messages received from service. Such clients call `Channel.receive()` and then
    process the returned message or sentinel.

    This protocol implementation handles only essential message processing:

    - Message parsing.
    - ACK-REQ for received STATE and NOOP messages.
    - Returns INVALID sentinel for received HELLO and CANCEL messages.
    - Closes the session when CLOSE message is received.
    """
    def __init__(self, *, session_type: Type[FBSPSession]=FBSPSession):
        """
        Arguments:
            session_type: Class for session objects.
        """
        super().__init__(session_type=session_type)
        self.handlers.update({MsgType.WELCOME: self.handle_welcome_msg,
                              MsgType.ERROR: self.handle_fbsp_msg,
                              MsgType.REPLY: self.handle_fbsp_msg,
                              MsgType.STATE: self.handle_fbsp_msg,
                              MsgType.NOOP: self.handle_noop_msg,
                              MsgType.DATA: self.handle_fbsp_msg,
                              MsgType.CLOSE: self.handle_close_msg,
                              MsgType.REQUEST: self.handle_fbsp_msg,
                              MsgType.HELLO: self.handle_invalid_msg,
                              MsgType.CANCEL: self.handle_invalid_msg,
                              })
        self._apis: Dict[ButlerInterface, int] = {}
    def handle_welcome_msg(self, channel: Channel, session: FBSPSession, msg: WelcomeMessage) -> WelcomeMessage:
        """Process WELCOME message received from service.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.
        """
        session.greeting = msg
        session.partner_uid = uuid.UUID(bytes=msg.data.service.uid)
        for intf in msg.data.api:
            self._apis[uuid.UUID(bytes=intf.uid)] = intf.number
        return msg
    def handle_fbsp_msg(self, channel: Channel, session: FBSPSession, msg: FBSPMessage) -> FBSPMessage:
        """FBSP message handler that simply returns received message.

        If message is a STATE message with ACK-REQ, sends ACK-REPLY. ACK-REQ for other
        messages must be handled by caller.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.

        Note:
            All exceptions are handled by `handle_exception`.
        """
        if (msg.msg_type is MsgType.STATE) and msg.has_ack_req():
            channel.send(self.create_ack_reply(msg), session)
        return msg
    def handle_invalid_msg(self, channel: Channel, session: FBSPSession, msg: FBSPMessage) -> FBSPMessage:
        """FBSP message handler that returns INVALID sentinel.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.
        """
        return INVALID
    def handle_noop_msg(self, channel: Channel, session: FBSPSession, msg: FBSPMessage) -> None:
        """Process NOOP message received from service.

        Sends ACK-REPLY if requested.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.

        Note:
            All exceptions are handled by `handle_exception`.
        """
        if msg.has_ack_req():
            channel.send(self.create_ack_reply(msg), session)
    def handle_close_msg(self, channel: Channel, session: FBSPSession, msg: FBSPMessage) -> FBSPMessage:
        """Process CLOSE message received from service.

        Discards the session and returns the CLOSE message received.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.
        """
        channel.discard_session(session)
        return msg
    def has_api(self, api: Type[ButlerInterface]) -> bool:
        """Returns True if attached service supports specified interface.

        Arguments:
            api: Service API (interface) enumeration.
        """
        return api.get_uid() in self._apis
    def create_request_for(self, session: FBSPSession, api_code: ButlerInterface, token: Token) -> APIMessage:
        """Returns new REQUEST message` for specific API call.

        Arguments:
            api_code: API Code
            token:    Message token
        """
        return self.create_message_for(MsgType.REQUEST, token,
                                       bb2h(self._apis[api_code.__class__.get_uid()],
                                            api_code.value))
    def exception_for(self, msg: ErrorMessage) -> ServiceError:
        """Returns `~saturnin.core.types.ServiceError` exception from ERROR message.
        """
        desc = [f"{msg.error_code.name}, relates to {msg.relates_to.name}"]
        for err in msg.errors:
            desc.append(f"#{err.code} : {err.description}")
        return ServiceError('\n'.join(desc))
    def send_hello(self, channel: Channel, session: FBSPSession, agent: AgentDescriptor,
                   peer: PeerDescriptor, token: Token=bytes(8)) -> None:
        """Sends HELLO message to the service.

        Arguments:
            channel: Channel used for connection to the service.
            session: Session associated with service.
            agent: Agent descriptor for client.
            peer: Peer descriptor for client.
            interfaces: List of service interface enumerations used by client.
            token: FBSP message token to be used in HELLO message.
        """
        msg: HelloMessage = self.create_message_for(MsgType.HELLO, token)
        msg.data.instance.uid = peer.uid.bytes
        msg.data.instance.pid = peer.pid
        msg.data.instance.host = peer.host
        msg.data.client.uid = agent.uid.bytes
        msg.data.client.name = agent.name
        msg.data.client.version = agent.version
        msg.data.client.classification = agent.classification
        msg.data.client.vendor.uid = agent.vendor_uid.bytes
        msg.data.client.platform.uid = agent.platform_uid.bytes
        msg.data.client.platform.version = agent.platform_version
        channel.send(msg, session)
    def send_close(self, channel: Channel, session: FBSPSession) -> None:
        """Sends `CLOSE` message to the service associated with session.

        Arguments:
            channel: Channel associate with data pipe.
            session: Session associated with transmission.
        """
        channel.send(self.create_message_for(MsgType.CLOSE, session.greeting.token), session)

class FBSPEventClient(FBSPClient):
    """4/FBSP - Firebird Butler Service Protocol - Client side.

    This is the EVENT version of client side FBSP protocol for clients that process
    FBSP messages received from service indirectly. Such clients use central I/O loop
    that process incomming messages in uniform way, and actual processing is done via event
    handlers.
    """
    def __init__(self, *, session_type: Type[FBSPSession]=FBSPSession):
        """
        Arguments:
            session_type: Class for session objects.
        """
        super().__init__(session_type=session_type)
        self.api_handlers: Dict[Any, Callable] = {}
        self.handlers.update({MsgType.WELCOME: self.handle_welcome_msg,
                              MsgType.ERROR: self.handle_error_msg,
                              MsgType.REPLY: self.handle_reply_msg,
                              MsgType.STATE: self.handle_state_msg,
                              MsgType.NOOP: self.handle_noop_msg,
                              MsgType.DATA: self.handle_data_msg,
                              MsgType.CLOSE: self.handle_close_msg,
                              MsgType.REQUEST: self.handle_ack_reply,
                              MsgType.HELLO: self.handle_invalid_msg,
                              MsgType.CANCEL: self.handle_invalid_msg,
                              })
    def register_api_handler(self, api_code: ButlerInterface, handler: Callable) -> None:
        """Register handler for REQUEST message for particular service API code.
        """
        _hndcheck.client_handler = handler
        _hndcheck.client_handler = None
        self.api_handlers[bb2h(self._apis[api_code.__class__.get_uid()], api_code.value)] = handler
    def handle_invalid_msg(self, channel: Channel, session: FBSPSession, msg: HelloMessage) -> None:
        """Process invalid message received from service.

        Raises `StopError` with error code `2 - Protocol violation`.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.

        Note:
            All exceptions are handled by `handle_exception`.
        """
        raise StopError("Unexpected message", code=ErrorCode.PROTOCOL_VIOLATION)
    def handle_welcome_msg(self, channel: Channel, session: FBSPSession, msg: WelcomeMessage) -> None:
        """Process WELCOME message received from service.

        Calls `on_service_connected` event handler.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.
        """
        super().handle_welcome_msg(channel, session, msg)
        self.on_service_connected(channel, session, msg, self)
    def handle_error_msg(self, channel: Channel, session: FBSPSession, msg: ErrorMessage) -> None:
        """Process ERROR message received from service.

        Calls `.on_error` event handler.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.

        Note:
            All exceptions are handled by `handle_exception`.
        """
        self.on_error(channel, session, msg)
    def handle_reply_msg(self, channel: Channel, session: FBSPSession, msg: APIMessage) -> None:
        """Process REPLY message received from service.

        Calls the appropriate API handler set via `.register_api_handler()`.

        Important:
            The event handler must send the ACK-REPLY message if requested. According to
            FBSP specification, the reply must be acknowledged without any delay, unless
            a previous agreement between the Client and the Service exists to handle it
            differently (for example when Client is prepared to accept subsequent DATA or
            other messages from Service).

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.

        Note:
            All exceptions are handled by `handle_exception`.
        """
        handler = self.api_handlers.get(msg.type_data)
        assert handler, "API REPLY without handler"
        handler(channel, session, msg, self)
    def handle_state_msg(self, channel: Channel, session: FBSPSession, msg: StateMessage) -> None:
        """Process STATE message received from service.

        Sends ACK-REPLY if requested and calls `.on_state` event handler.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.

        Note:
            All exceptions are handled by `handle_exception`.
        """
        if msg.has_ack_req():
            channel.send(self.create_ack_reply(msg), session)
        self.on_state(channel, session, msg)
    def handle_noop_msg(self, channel: Channel, session: FBSPSession, msg: FBSPMessage) -> None:
        """Process NOOP message received from service.

        If message is an ACK-REPLY, it calls `.on_ack_received` event handler.
        Otherwise it sends ACK-REPLY if requested and then calls `.on_noop` event handler.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.

        Note:
            All exceptions are handled by `handle_exception`.
        """
        if msg.has_ack_reply():
            self.on_ack_received(channel, session, msg)
            return
        if msg.has_ack_req():
            channel.send(self.create_ack_reply(msg), session)
        self.on_noop(channel, session)
    def handle_data_msg(self, channel: Channel, session: FBSPSession, msg: DataMessage) -> None:
        """Process DATA message received from peer.

        If message is an ACK-REPLY, it calls `.on_ack_received` event handler.
        Otherwise it calls `.on_data` event handler that must handle the ACK-REQ if set.
        If `.on_data` event handler is not defined, the ACK-REPLY message is send by this
        handler.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.

        Note:
            All exceptions are handled by `handle_exception`.
        """
        if msg.has_ack_reply():
            self.on_ack_received(channel, session, msg)
            return
        self.on_data(channel, session, msg)
        if not self.on_data.is_set() and msg.has_ack_req():
            channel.send(self.create_ack_reply(msg), session)
    def handle_close_msg(self, channel: Channel, session: FBSPSession, msg: FBSPMessage) -> None:
        """Process CLOSE message received from service.

        Calls `.on_session_closed` event handler and then discards the session.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.
        """
        try:
            self.on_session_closed(channel, session, msg)
        except:
            # We don't want to handle this via `handle_exception` and we're closing
            # the session anyway
            pass
        finally:
            channel.discard_session(session)
    def handle_ack_reply(self, channel: Channel, session: FBSPSession, msg: FBSPMessage) -> None:
        """Process ACK-REPLY REQUEST messages received from service.

        If message has ACK-REPLY flag set, it calls `.on_ack_received` event handler.
        Otherwise it's ignored because it violates the protocol.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received REPLY or STATE message.

        Note:
            All exceptions are handled by `handle_exception`.
        """
        if msg.has_ack_reply():
            self.on_ack_received(channel, session, msg)
    @eventsocket
    def on_service_connected(self, channel: Channel, session: FBSPSession,
                             msg: WelcomeMessage, protocol: FBSPEventClient) -> None:
        """Called when WELCOME message is received from service.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.

        The event handler should register all service API handlers for REPLY messages
        (see `FBSPEventClient.register_api_handler()`).
        """

    @eventsocket
    def on_noop(self, channel: Channel, session: FBSPSession) -> None:
        """Called when NOOP message is received, and after ACK-REPLY (if requested) is send.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with peer.
        """
    @eventsocket
    def on_data(self, channel: Channel, session: FBSPSession, msg: DataMessage) -> None:
        """Called when DATA message is received.

        Important:
            If handler is defined, it must send the ACK-REPLY if received message has
            ACK-REQ set. It's up to service interface whether ACK-REPLY would be sent
            before, or after data contained within message are processed.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.

        Note:
            All exceptions are handled by `handle_exception`.
        """
    @eventsocket
    def on_error(self, channel: Channel, session: FBSPSession, msg: ErrorMessage) -> None:
        """Called when ERROR message is received.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.

        Note:
            All exceptions are handled by `handle_exception`.
        """
    @eventsocket
    def on_state(self, channel: Channel, session: FBSPSession, msg: StateMessage) -> None:
        """Called when STATE message is received.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.

        Note:
            All exceptions are handled by `handle_exception`.
        """
    @eventsocket
    def on_session_closed(self, channel: Channel, session: FBSPSession, msg: FBSPMessage,
                          exc: Exception=None) -> None:
        """Called when CLOSE message is received or sent, to release any resources
        associated with current transmission.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with peer.
            msg: Received/sent CLOSE message.
            exc: Exception that caused the error.
        """
    @eventsocket
    def on_ack_received(self, channel: Channel, session: FBSPSession, msg: DataMessage) -> None:
        """Called when ACK-REPLY NOOP, DATA or REQUEST message is received.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received ACK-REPLY message.

        Note:
            All exceptions are handled by `handle_exception`.
        """

class __APIHandlerChecker:
    @eventsocket
    def service_handler(self, channel: Channel, session: FBSPSession, msg: FBSPMessage,
                        protocol: FBSPService) -> None:
        ""
    @eventsocket
    def client_handler(self, channel: Channel, session: FBSPSession, msg: FBSPMessage,
                       protocol: FBSPEventClient) -> None:
        ""

_hndcheck: __APIHandlerChecker = __APIHandlerChecker()



#coding:utf-8
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/component/registry.py
# DESCRIPTION:    Component registration and discovery
# CREATED:        4.12.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2020 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Saturnin component registration and discovery


"""

from __future__ import annotations
from typing import List
from pkg_resources import iter_entry_points
from functools import partial
from itertools import chain
from saturnin.base import ServiceDescriptor, Error

_iterators = [partial(iter_entry_points, 'saturnin.service')]

for i in iter_entry_points('saturnin.service.iterator'):
    _iterators.append(i.load())

def get_service_desciptors(uid: str=None) -> List[ServiceDescriptor]:
    """Returns list of service descriptors for registered services.
    """
    result = []
    for e in chain.from_iterable([i(uid) for i in _iterators]):
        try:
            result.append(e.load())
        except Exception as exc:
            raise Error(f"Descriptor loading failed: {str(e)}") from exc
    return result



#coding:utf-8
#
# PROGRAM/MODULE: saturnin-core
# FILE:           saturnin/core/__init__.py
# DESCRIPTION:    Saturnin (Firebird Butler Development Platform) core package
# CREATED:        21.2.2019
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2019 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________.

"Saturnin (Firebird Butler Development Platform) core package"

import uuid
from firebird.base.types import Error, ZMQAddress, DEFAULT, UNDEFINED, ANY, load
from firebird.base.config import Config, ConfigProto
from .types import PLATFORM_OID, PLATFORM_UID, PLATFORM_VERSION, VENDOR_OID, VENDOR_UID, \
     RoutingID, Token, TSupplement, INVALID, TIMEOUT, \
     InvalidMessageError, ChannelError, ServiceError, ClientError, StopError, \
     Origin, SocketMode, Direction, SocketType, State,  PipeSocket, FileOpenMode, Outcome, \
     ButlerInterface, AgentDescriptor, PeerDescriptor, ServiceDescriptor, PrioritizedItem, \
     MIME, MIME_TYPE_PROTO, MIME_TYPE_TEXT, PROTO_PEER
from .transport import ChannelManager, Channel, Message, SimpleMessage, Protocol, Session, \
     DealerChannel, RouterChannel, PushChannel, PullChannel, PubChannel, SubChannel, \
     XPubChannel, XSubChannel, PairChannel, \
     TZMQMessage, TMessageHandler, TSocketOptions, INTERNAL_ROUTE
from .component import Component, ComponentConfig
from .site import site

def pkg_name(file: str) -> str:
    """Returns name of the package to which the `file` belongs.
    """
    return __import__(file).__package__



#coding:utf-8
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/lib/data/filter.py
# DESCRIPTION:    Base class for Saturnin data filter microservices
# CREATED:        14.12.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2020 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Saturnin base class for data filter microservices
"""

from __future__ import annotations
from typing import Any, List, cast
from functools import partial
from collections import deque
from firebird.base.config import MIME, StrOption, EnumOption, \
     IntOption, BoolOption, ZMQAddressOption, MIMEOption
from saturnin.base import Error, StopError, Direction, SocketMode, PipeSocket, Outcome, \
     ZMQAddress, MIME, ComponentConfig, Message, SimpleMessage, Session, Protocol, Channel, \
     DealerChannel, PushChannel, PullChannel, ANY
from saturnin.component.micro import MicroService
from saturnin.protocol.fbdp import ErrorCode, FBDPServer, FBDPClient, \
     FBDPSession, FBDPMessage

INPUT_PIPE_CHN = 'input-pipe'
OUTPUT_PIPE_CHN = 'output-pipe'
WAKE_PUSH_CHN = 'wake-push'
WAKE_PULL_CHN = 'wake-pull'

class DataFilterConfig(ComponentConfig):
    """Base data provider microservice configuration."""
    def __init__(self, name: str):
        super().__init__(name)
        self.propagate_input_error: BoolOption = \
            BoolOption('propagate_input_error',
                       "When input pipe is closed with error, close output with error as well",
                       default=True)
        # Input pipe
        self.input_pipe: StrOption = \
            StrOption('input_pipe', "Data Pipe Identification", required=True)
        self.input_pipe_address: ZMQAddressOption = \
            ZMQAddressOption('input_pipe_address', "Data Pipe endpoint address", required=True)
        self.input_pipe_mode: EnumOption = \
            EnumOption('input_pipe_mode', SocketMode, "Data Pipe Mode", required=True)
        self.input_pipe_format: MIMEOption = \
            MIMEOption('input_pipe_format', "Pipe data format specification")
        self.input_batch_size: IntOption = \
            IntOption('input_batch_size', "Data batch size", required=True, default=50)
        self.input_ready_schedule_interval: IntOption = \
            IntOption('input_ready_schedule_interval',
                      "READY message schedule interval in milliseconds", required=True,
                      default=1000)
        # Output pipe
        self.output_pipe: StrOption = \
            StrOption('output_pipe', "Data Pipe Identification", required=True)
        self.output_pipe_address: ZMQAddressOption = \
            ZMQAddressOption('output_pipe_address', "Data Pipe endpoint address", required=True)
        self.output_pipe_mode: EnumOption = \
            EnumOption('output_pipe_mode', SocketMode, "Data Pipe Mode", required=True)
        self.output_pipe_format: MIMEOption = \
            MIMEOption('output_pipe_format', "Pipe data format specification")
        self.output_batch_size: IntOption = \
            IntOption('output_batch_size', "Data batch size", required=True, default=50)
        self.output_ready_schedule_interval: IntOption = \
            IntOption('output_ready_schedule_interval',
                      "READY message schedule interval in milliseconds", required=True,
                      default=1000)
    def validate(self) -> None:
        """Extended validation.

        - `pipe_format` is required for CONNECT `pipe_mode`.
        """
        super().validate()
        if self.input_pipe_mode.value is SocketMode.CONNECT and self.input_pipe_format.value is None:
            raise Error("'input_pipe_format' required for CONNECT pipe mode.")
        if self.output_pipe_mode.value is SocketMode.CONNECT and self.output_pipe_format.value is None:
            raise Error("'output_pipe_format' required for CONNECT pipe mode.")

class DataFilterMicro(MicroService):
    """Base data provider microservice.

    Descendant classes should override:

    - `handle_input_accept_client` to validate client request and aquire resources
      associated with input pipe.
    - `handle_output_accept_client` to validate client request and aquire resources
      associated with output pipe.
    - `handle_output_produce_data` to produce data for outgoing DATA message.
    - `handle_input_accept_data` to process received data. CONSUMER only.
    - `handle_input_pipe_closed` to release resource assiciated with input pipe.
    - `handle_output_pipe_closed` to release resource assiciated with output pipe.
    """
    def initialize(self, config: DataFilterConfig) -> None:
        """Verify configuration and assemble component structural parts.
        """
        super().initialize(config)
        # Closing flag
        self.closing: bool = False
        # Configuration
        self.propagate_input_error = config.propagate_input_error.value
        # INPUT pipe
        self.input_pipe: str = config.input_pipe.value
        self.input_pipe_mode: SocketMode = config.input_pipe_mode.value
        self.input_pipe_address: ZMQAddress = config.input_pipe_address.value
        self.input_pipe_format: MIME = config.input_pipe_format.value
        self.input_batch_size: int = config.input_batch_size.value
        self.input_ready_schedule_interval: int = config.input_ready_schedule_interval.value
        #
        if self.input_pipe_mode == SocketMode.BIND:
            self.input_protocol = FBDPServer()
            self.input_protocol.on_exception = self.handle_exception
            self.input_protocol.on_accept_client = self.handle_input_accept_client
            self.input_protocol.on_schedule_ready = self.handle_input_schedule_ready
            # We have an endpoint to bind
            self.endpoints[INPUT_PIPE_CHN] = [self.input_pipe_address]
        else:
            self.input_protocol = FBDPClient()
        self.input_protocol.batch_size = self.input_batch_size
        self.input_protocol.on_pipe_closed = self.handle_input_pipe_closed
        self.input_protocol.on_accept_data = self.handle_input_accept_data
        self.input_protocol.on_get_data = self.handle_input_get_data
        # Create INPUT pipe channel
        self.pipe_in_chn: Channel = self.mngr.create_channel(DealerChannel, INPUT_PIPE_CHN,
                                                             self.input_protocol,
                                                             wait_for=Direction.IN)
        self.pipe_in_chn.protocol.log_context = self.logging_id
        # OUTPUT pipe
        self.output_pipe: str = config.output_pipe.value
        self.output_pipe_mode: SocketMode = config.output_pipe_mode.value
        self.output_pipe_address: ZMQAddress = config.output_pipe_address.value
        self.output_pipe_format: MIME = config.output_pipe_format.value
        self.output_batch_size: int = config.output_batch_size.value
        self.output_ready_schedule_interval: int = config.output_ready_schedule_interval.value
        #
        if self.output_pipe_mode == SocketMode.BIND:
            self.output_protocol = FBDPServer()
            self.output_protocol.on_exception = self.handle_exception
            self.output_protocol.on_accept_client = self.handle_output_accept_client
            self.output_protocol.on_schedule_ready = self.handle_output_schedule_ready
            # We have an endpoint to bind
            self.endpoints[OUTPUT_PIPE_CHN] = [self.output_pipe_address]
        else:
            self.output_protocol = FBDPClient()
        self.output_protocol.batch_size = self.output_batch_size
        self.output_protocol.on_pipe_closed = self.handle_output_pipe_closed
        self.output_protocol.on_produce_data = self.handle_output_produce_data
        self.output_protocol.on_get_data = self.handle_output_get_data
        # Create OUTPUT pipe channel
        self.pipe_out_chn: Channel = self.mngr.create_channel(DealerChannel, OUTPUT_PIPE_CHN,
                                                              self.output_protocol,
                                                              wait_for=Direction.IN)
        self.pipe_out_chn.protocol.log_context = self.logging_id
        # Awake channels
        self.wake_address: ZMQAddress = ZMQAddress(f'inproc://{self.peer.uid.hex}-wake')
        wake_protocol = Protocol()
        wake_protocol.handlers[ANY] = self.handle_wake_msg
        # PUSH wake
        self.wake_out_chn: Channel = self.mngr.create_channel(PushChannel, WAKE_PUSH_CHN,
                                                              wake_protocol)
        self.wake_out_chn.protocol.log_context = self.logging_id
        # PULL wake
        self.wake_in_chn: Channel = self.mngr.create_channel(PullChannel, WAKE_PULL_CHN,
                                                             wake_protocol,
                                                             wait_for=Direction.IN)
        self.wake_in_chn.protocol.log_context = self.logging_id
        # We have an endpoint to bind
        self.endpoints[WAKE_PULL_CHN] = [self.wake_address]
        #: Data to be sent to output.
        self.output: deque = deque()
    def aquire_resources(self) -> None:
        """Aquire resources required by component (open files, connect to other services etc.).

        Must raise an exception when resource aquisition fails.
        """
        # Connect wake PUSH
        self.wake_out_chn.connect(self.wake_address)
        # Connect to the data pipes
        # INPUT pipe
        if self.input_pipe_mode == SocketMode.CONNECT:
            session = self.pipe_in_chn.connect(self.input_pipe_address)
            # OPEN the data pipe connection, this also fills session attributes
            # We are CONSUMER client, we must attach to server OUTPUT
            cast(FBDPClient, self.pipe_in_chn.protocol).send_open(self.pipe_in_chn,
                                                                  session, self.input_pipe,
                                                                  PipeSocket.OUTPUT,
                                                                  self.input_pipe_format)
        # OUTPUT pipe
        if self.output_pipe_mode == SocketMode.CONNECT:
            session = self.pipe_out_chn.connect(self.output_pipe_address)
            # OPEN the data pipe connection, this also fills session attributes
            # We are PRODUCER client, we must attach to server INPUT
            cast(FBDPClient, self.pipe_out_chn.protocol).send_open(self.pipe_out_chn,
                                                                   session, self.output_pipe,
                                                                   PipeSocket.INPUT,
                                                                   self.output_pipe_format)
    def release_resources(self) -> None:
        """Release resources aquired by component (close files, disconnect from other services etc.)
        """
        # Disonnect wake PUSH
        for session in list(self.wake_out_chn.sessions.values()):
            self.wake_out_chn.discard_session(session)
        # CLOSE all active data input pipe sessions
        # send_close() will discard session, so we can't iterate over sessions.values() directly
        for session in list(self.pipe_in_chn.sessions.values()):
            # We have to report error here, because normal is to close pipes before
            # shutdown is commenced. Mind that service shutdown could be also caused by error!
            cast(FBDPServer, self.pipe_in_chn.protocol).send_close(self.pipe_in_chn, session, ErrorCode.ERROR)
        # CLOSE all active data output pipe sessions
        # send_close() will discard session, so we can't iterate over sessions.values() directly
        for session in list(self.pipe_out_chn.sessions.values()):
            # We have to report error here, because normal is to close pipes before
            # shutdown is commenced. Mind that service shutdown could be also caused by error!
            cast(FBDPServer, self.pipe_out_chn.protocol).send_close(self.pipe_out_chn,
                                                                    session, ErrorCode.ERROR)
    def store_output(self, data: Any) -> None:
        """Store data to output queue and send wake notification.
        """
        self.output.append(data)
        msg = SimpleMessage()
        msg.data.append(b'wake')
        self.wake_out_chn.send(msg, self.wake_out_chn.session)
    def store_batch_output(self, batch: List) -> None:
        """Store batch of data to output queue and send wake notification.
        """
        for data in batch:
            self.output.append(data)
        if self.output:
            msg = SimpleMessage()
            msg.data.append(b'wake')
            self.wake_out_chn.send(msg, self.wake_out_chn.session)
    def finish_input_processing(self, channel: Channel, session: FBDPSession, code: ErrorCode) -> None:
        """Called when input pipe is closed while output pipe will remain open.

        When code is ErrorCode.OK, the input was closed normally. Otherwise it indicates
        the type of problem that caused the input to be closed.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with peer.
            code:    Input pipe closing ErrorCode.

        Note:
            The default implementation does nothing.
        """
    def handle_wake_msg(self, channel: Channel, session: Session, msg: Message) -> None:
        """Handler for "data available" pings sent via wake channels.
        """
        if not self.output:
            # Unlikely case when we've got wake but all data were already sent
            return
        if not self.pipe_out_chn.sessions:
            # We need active pipe connection
            return
        session: FBDPSession = self.pipe_out_chn.session
        if session.transmit is not None:
            # Transmission in progress, make sure that we will send data
            self.pipe_out_chn.set_wait_out(True, session)
        elif self.output_pipe_mode is SocketMode.BIND and not session.await_ready:
            # We are server without active transmission and READY was not sent yet, so we
            # can send READY immediately
            cast(FBDPServer, self.pipe_out_chn.protocol)._init_new_batch(self.pipe_out_chn,
                                                                         session)
    def handle_exception(self, channel: Channel, session: Session, msg: Message,
                         exc: Exception) -> None:
        """Event handler called by `.handle_msg()` on exception in message handler.

        Sets service outcome to ERROR and notes exception as details.
        """
        if isinstance(exc, StopError):
            if getattr(exc, 'code', None) is ErrorCode.OK:
                return
        self.outcome = Outcome.ERROR
        self.details = exc
    def handle_input_get_data(self, channel: Channel, session: FBDPSession) -> bool:
        """Event handler executed to query the service whether is ready to accept input data.

        Returns True if output pipe is open, otherwise false.
        """
        return bool(self.pipe_in_chn.sessions)
    def handle_output_get_data(self, channel: Channel, session: FBDPSession) -> bool:
        """Event handler executed to query the data source for data availability.

        Returns True if output deque contains any data.

        Cancels the transmission by raising the `StopError` if there are no output data
        and input pipe is closed.
        """
        have_data = bool(self.output)
        if not have_data and not self.pipe_in_chn.sessions:
            raise StopError("EOF", code=ErrorCode.OK)
        return have_data
    # FBDP server only
    def handle_input_accept_client(self, channel: Channel, session: FBDPSession) -> None:
        """Event handler executed when client connects to INPUT data pipe via OPEN message.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with client.

        The session attributes `data_pipe`, `pipe_socket`, `data_format` and `params`
        contain information sent by client, and the event handler validates the request.

        If request should be rejected, it raises the `StopError` exception with `code`
        attribute containing the `ErrorCode` to be returned in CLOSE message.

        Important:
            Base implementation validates pipe identification and pipe socket, and converts
            data format from string to MIME (in session).

            The descendant class that overrides this method must call super() as first
            action.
        """
        if session.pipe != self.output_pipe:
            raise StopError(f"Unknown data pipe '{session.data_pipe}'",
                            code = ErrorCode.PIPE_ENDPOINT_UNAVAILABLE)
        # Clients can attach only to INPUT
        elif session.socket is not PipeSocket.INPUT:
            raise StopError(f"'{session.socket}' socket not available",
                            code = ErrorCode.PIPE_ENDPOINT_UNAVAILABLE)
        # We work with MIME formats, so we'll convert the format specification to MIME
        session.data_format = MIME(session.data_format)
    def handle_output_accept_client(self, channel: Channel, session: FBDPSession) -> None:
        """Event handler executed when client connects to OUTPUT data pipe via OPEN message.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with client.

        The session attributes `data_pipe`, `pipe_socket`, `data_format` and `params`
        contain information sent by client, and the event handler validates the request.

        If request should be rejected, it raises the `StopError` exception with `code`
        attribute containing the `ErrorCode` to be returned in CLOSE message.

        Important:
            Base implementation validates pipe identification and pipe socket, and converts
            data format from string to MIME (in session).

            The descendant class that overrides this method must call super() as first
            action.
        """
        if session.pipe != self.output_pipe:
            raise StopError(f"Unknown data pipe '{session.data_pipe}'",
                            code = ErrorCode.PIPE_ENDPOINT_UNAVAILABLE)
        # Clients can attach only to our OUTPUT
        elif session.socket is not PipeSocket.OUTPUT:
            raise StopError(f"'{session.socket}' socket not available",
                            code = ErrorCode.PIPE_ENDPOINT_UNAVAILABLE)
        # We work with MIME formats, so we'll convert the format specification to MIME
        session.data_format = MIME(session.data_format)
    def handle_input_schedule_ready(self, channel: Channel, session: FBDPSession) -> None:
        """Event handler executed in order to send the READY message to the client later.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with client.

        The event handler may cancel the transmission by raising the `StopError` exception
        with `code` attribute containing the `ErrorCode` to be returned in CLOSE message.

        Important:
            The base implementation schedules `resend_ready()` according to
            `ready_schedule_interval` configuration option.
        """
        self.schedule(partial(cast(FBDPServer, channel.protocol).resend_ready,
                              channel, session),
                      self.input_ready_schedule_interval)
    def handle_output_schedule_ready(self, channel: Channel, session: FBDPSession) -> None:
        """Event handler executed in order to send the READY message to the client later.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with client.

        The event handler may cancel the transmission by raising the `StopError` exception
        with `code` attribute containing the `ErrorCode` to be returned in CLOSE message.

        Important:
            The base implementation schedules `resend_ready()` according to
            `ready_schedule_interval` configuration option.
        """
        self.schedule(partial(cast(FBDPServer, channel.protocol).resend_ready,
                              channel, session),
                      self.output_ready_schedule_interval)
    # FBDP common
    def handle_output_produce_data(self, channel: Channel, session: FBDPSession, msg: FBDPMessage) -> None:
        """Event handler executed to store data into outgoing DATA message.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with client.
            msg: DATA message that will be sent to client.

        Important:
            The base implementation simply raises StopError with ErrorCode.OK code, so
            the descendant class must override this method without super() call.

            The event handler must `popleft()` data from `output` queue and store them in
            `msg.data_frame` attribute. It may also set ACK-REQUEST flag and `type_data`
            attribute.

            The event handler may cancel the transmission by raising the `StopError`
            exception with `code` attribute containing the `ErrorCode` to be returned in
            CLOSE message.

        Note:
            To indicate end of data, raise StopError with ErrorCode.OK code.
        """
        raise StopError('OK', code=ErrorCode.OK)
    def handle_input_accept_data(self, channel: Channel, session: FBDPSession, data: bytes) -> None:
        """Event handler executed to process data received in DATA message.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with client.
            data: Data received from client.

        Important:
            Any output data produced by event handler must be stored into output queue via
            `store_output()` method.

            The base implementation simply raises StopError with ErrorCode.OK code, so
            the descendant class must override this method without super() call.

            The event handler may cancel the transmission by raising the `StopError`
            exception with `code` attribute containing the `ErrorCode` to be returned in
            CLOSE message.

        Note:
            The ACK-REQUEST in received DATA message is handled automatically by protocol.
        """
        raise StopError('OK', code=ErrorCode.OK)
    def handle_input_pipe_closed(self, channel: Channel, session: FBDPSession, msg: FBDPMessage,
                                 exc: Exception=None) -> None:
        """Event handler executed when CLOSE message is received or sent, to release any
        resources associated with current transmission.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with peer.
            msg: Received/sent CLOSE message.
            exc: Exception that caused the error.

        Important:
            The base implementation does next actions:

            - If exception is provided, sets service execution outcome to ERROR
              and notes exception in details.
            - Closes the output pipe it's still open  and `closing` flag is False, and if
              the input is not closed normally and `propagate_input_error` is True.
            - Sets the signal to stop the service.

            The descendant class that overrides this method must call super().
        """
        # FDBP converts exceptions raised in our event handler to CLOSE messages, so
        # here is the central place to handle errors in data pipe processing.
        code: ErrorCode = msg.type_data
        if exc is not None:
            # Note problem in service execution outcome
            if code is not ErrorCode.OK:
                self.outcome = Outcome.ERROR
                self.details = exc
        self.finish_input_processing(channel, session, code)
        # If input is not closed normally and we should propagate the problem, or when
        # there is no data for output, close the output pipe if it's still open
        if (code is not ErrorCode.OK and self.propagate_input_error) or not self.output:
            if not self.closing:
                self.closing = True
                for session in self.pipe_out_chn.sessions:
                    cast(FBDPServer, self.pipe_out_chn.protocol).send_close(self.pipe_out_chn,
                                                                            session, code, exc)
            # Request service to stop
            self.stop.set()
        self.closing = False
    def handle_output_pipe_closed(self, channel: Channel, session: FBDPSession, msg: FBDPMessage,
                                  exc: Exception=None) -> None:
        """Event handler executed when CLOSE message is received or sent, to release any
        resources associated with current transmission.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with peer.
            msg: Received/sent CLOSE message.
            exc: Exception that caused the error.

        Important:
            The base implementation does next actions:

            - If exception is provided, sets service execution outcome to ERROR
              and notes exception in details.
            - Closes the input pipe if it's still open and `closing` flag is False.
            - Sets the signal to stop the service.

            The descendant class that overrides this method must call super().
        """
        # FDBP converts exceptions raised in our event handler to CLOSE messages, so
        # here is the central place to handle errors in data pipe processing.
        code: ErrorCode = msg.type_data
        if exc is not None:
            # Note problem in service execution outcome
            if code is not ErrorCode.OK:
                self.outcome = Outcome.ERROR
                self.details = exc
        # Close the input pipe if it's still open
        if not self.closing:
            self.closing = True
            for session in self.pipe_in_chn.sessions:
                cast(FBDPServer, self.pipe_in_chn.protocol).send_close(self.pipe_in_chn,
                                                                       session, code, exc)
        # Request service to stop
        self.stop.set()
        self.closing = False



#coding:utf-8
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/lib/pyproject.py
# DESCRIPTION:    Class to handle pyproject.toml files.
# CREATED:        12.2.2021
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2021 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""saturnin - Class to handle pyproject.toml files.


"""

from __future__ import annotations
from typing import List
import re

def normalize(name):
    return re.sub(r"[-_.]+", "-", name).lower()


class PyProjectTOML:
    """
    """
    def __init__(self, pyproject_data: Dict):
        project = pyproject_data['project']
        self.name: str = normalize(project['name'])
        self.original_name: str = project['name']
        self.version: str = project['version']
        self.description: str = project['description']
        self.readme: Union[str, Dict[str, str]] = project.get('readme')
        self.requires_python: str = project.get('requires-python')
        self.license: Dict[str, str] = project.get('license')
        self.authors: List[Dict[str, str]] = project.get('authors')
        self.maintainers: List[Dict[str, str]] = project.get('keywords')
        self.keywords: List[str] = project.get('keywords')
        self.classifiers: List[str] = project.get('classifiers')
        self.urls: Dict[str, str] = project.get('urls')
        self.scripts: Dict[str, str] = project.get('scripts')
        self.gui_scripts: Dict[str, str] = project.get('gui-scripts')
        self.entry_points: Dict[str, Dict[str, str]] = project.get('entry-points')
        self.dependencies: List[str] = project.get('dependencies')
        self.optional_dependencies: Dict[str, List[str]] = project.get('optional-dependencies')
        self.dynamic: List[str] = project.get('dynamic')
        #
        meta = pyproject_data['tool']['saturnin']['metadata']
        self.component_type: str = meta['component-type']
        self.uid: str = meta['uid']
        self.descriptor: str = meta['descriptor']
    def make_setup_cfg(self, cfg: 'ConfigParser') -> None:
        """
        """
        def set_content_type(readme_file: str):
            if readme_file.lower().endswith('.md'):
                cfg['metadata']['long_description_content_type'] = 'text/markdown; charset=UTF-8'
            elif readme_file.lower().endswith('.rst'):
                cfg['metadata']['long_description_content_type'] = 'text/x-rst; charset=UTF-8'
            else:
                cfg['metadata']['long_description_content_type'] = 'text/plain; charset=UTF-8'
        def multiline(lines: List) -> str:
            l = ['']
            l.extend(lines)
            return '\n'.join(l)

        if not cfg.has_section('metadata'):
            cfg.add_section('metadata')
        cfg['metadata']['name'] = self.name
        cfg['metadata']['version'] = self.version
        cfg['metadata']['description'] = self.description
        cfg['metadata']['name'] = self.name
        if self.readme is not None:
            if isinstance(self.readme, dict):
                if 'file' in self.readme:
                    cfg['metadata']['long_description'] = f"file: {self.readme['file']}"
                else:
                    cfg['metadata']['long_description'] = self.readme['text']
                if 'content-type' in self.readme:
                    cfg['metadata']['long_description_content_type'] = self.readme['content-type']
                elif 'file' in self.readme:
                    set_content_type(self.readme['file'])
            else:
                cfg['metadata']['long_description'] = f"file: {self.readme}"
                set_content_type(self.readme)
        if self.requires_python is not None:
            cfg['metadata']['python_requires'] = self.requires_python
        if self.license is not None:
            if 'file' in self.license:
                cfg['metadata']['license_file'] = self.license['file']
            else:
                cfg['metadata']['license'] = self.license['text']
        if self.authors is not None:
            author = self.authors[0]
            if 'name' in author:
                cfg['metadata']['author'] = author['name']
            if 'email' in author:
                cfg['metadata']['author_email'] = author['email']
            if 'name' in author and 'email' in author:
                cfg['metadata']['author_email'] = f"{author['name']} <{author['email']}>"
        if self.maintainers is not None:
            maintainer = self.maintainers[0]
            if 'name' in maintainer:
                cfg['metadata']['maintainer'] = maintainer['name']
            if 'email' in maintainer:
                cfg['metadata']['maintainer_email'] = maintainer['email']
            if 'name' in maintainer and 'email' in maintainer:
                cfg['metadata']['maintainer_email'] = f"{maintainer['name']} <{maintainer['email']}>"
        if self.keywords is not None:
            cfg['metadata']['keywords'] = ', '.join(self.keywords)
        if self.classifiers is not None:
            cfg['metadata']['classifiers'] = multiline(self.classifiers)
        if self.urls is not None:
            urls = []
            for k, v in self.urls.items():
                if k.lower() == 'home':
                    cfg['metadata']['url'] = v
                else:
                    urls.append(f'{k} = {v}')
            cfg['metadata']['project_urls '] = multiline(urls)
        if self.scripts is not None:
            if not cfg.has_section('options.entry_points'):
                cfg.add_section('options.entry_points')
            cfg['options.entry_points']['console_scripts'] = \
                multiline([f'{k} = {v}' for k, v in self.scripts.items()])
        if self.gui_scripts is not None:
            if not cfg.has_section('options.entry_points'):
                cfg.add_section('options.entry_points')
            cfg['options.entry_points']['gui_scripts'] = \
                multiline([f'{k} = {v}' for k, v in self.gui_scripts.items()])
        if self.entry_points is not None:
            if not cfg.has_section('options.entry_points'):
                cfg.add_section('options.entry_points')
            for group_name, group in self.entry_points.items():
                cfg['options.entry_points'][group_name] = \
                    multiline([f'{k} = {v}' for k, v in group.items()])
        if self.dependencies is not None:
            if not cfg.has_section('options'):
                cfg.add_section('options')
            cfg['options']['install_requires'] = multiline(self.dependencies)
        if self.optional_dependencies is not None:
            if not cfg.has_section('options.extras_require'):
                cfg.add_section('options.extras_require')
            for ext_name, extra in self.optional_dependencies.items():
                cfg['options.extras_require'][ext_name] = \
                    multiline([f'{k} = {v}' for k, v in extra.items()])




#coding:utf-8
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/transport.py
# DESCRIPTION:    ZeroMQ messaging - base classes and other definitions
# CREATED:        28.2.2019
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2019 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________.

"Saturnin ZeroMQ messaging - base classes and other definitions"

from __future__ import annotations
from typing import  Union, Dict, List, Iterable, Callable, Optional, Type, Any
from abc import ABC, abstractmethod
from weakref import proxy
import uuid
import warnings
import zmq
from zmq import Frame, ZMQError, Again, POLLIN, POLLOUT
from firebird.base.types import ZMQAddress, DEFAULT, UNDEFINED, ANY
from firebird.base.signal import eventsocket
from firebird.base.logging import LoggingIdMixin
from firebird.base.trace import TracedMixin
from .types import RoutingID, SocketMode, SocketType, Direction, \
     InvalidMessageError, ChannelError, INVALID, TIMEOUT

#: Internal routing ID
INTERNAL_ROUTE: RoutingID = b'INTERNAL'

class ChannelManager(LoggingIdMixin, TracedMixin):
    """Manager of ZeroMQ communication channels.
    """
    def __init__(self, context: zmq.Context):
        """
        Arguments:
            context: ZMQ Context instance.
        """
        #: ZMQ Context instance.
        self.ctx: zmq.Context = context
        #: Dictionary with managed channels. Key is `Channel.name`, value is the `Channel`.
        self.channels: Dict[str, Channel] = {}
        #: Logging context
        self.log_context: Any = UNDEFINED
        self._poller: zmq.Poller = None
        self._chmap: Dict[zmq.Socket, Channel] = {}
        self._pollout: bool = False
    def create_channel(self, cls: Type[Channel], name: str, protocol: Protocol, *,
                       routing_id: RoutingID=DEFAULT, session_type: Type[Session]=DEFAULT,
                       wait_for: Direction=Direction.NONE,
                       snd_timeout: int=100, rcv_timeout: int=100,
                       linger: int=5000, sock_opts: TSocketOptions=None) -> Channel:
        """Creates new channel.

        Arguments:
            cls: Channel class.
            name: Channel name.
            routing_id: Channel socket identity (routing ID for peers).
            protocol: Protocol for serializing/deserializing messages.
            session_type: Session type. DEFAULT session type is obtained from MessageHandler.
            wait_for: Direction(s) of transmission events for this channel processed by `.wait()`.
            snd_timeout: Timeout for send operation in milliseconds, None means infinite.
            rcv_timeout: Timeout for receive operation in milliseconds, None means infinite.
            linger: ZMQ socket linger period.
            sock_opts: Dictionary with socket additional options.
        """
        assert name not in self.channels, "Duplicate channel name"
        chn: Channel = cls(self, name, protocol, routing_id, session_type, wait_for,
                           snd_timeout, rcv_timeout, linger, sock_opts)
        self.channels[chn.name] = chn
        return chn
    def update_poller(self, channel: Channel, value: Direction) -> None:
        """Update poller registration for channel.
        """
        self._pollout = False
        for chn in self.channels.values():
            self._pollout = self._pollout or Direction.OUT in chn.wait_for
        if self._poller is not None:
            self._poller.modify(channel.socket, value.value)
    def has_pollout(self) -> bool:
        """Returns True if :meth:`wait` will check for POLLOUT event on any channel.
        """
        return self._pollout
    def wait(self, timeout: int=None) -> Dict[Channel, Direction]:
        """Wait for I/O events on channnels.

        Arguments:
            timeout: The timeout in milliseconds. `None` value means `infinite`.

        Returns:
            Dictionary with channel keys and event values.
        """
        if self._poller is None:
            self._poller = zmq.Poller()
            self._pollout = False
            for chn in self.channels.values():
                self._pollout = self._pollout or Direction.OUT in chn.wait_for
                self._poller.modify(chn.socket, chn.wait_for.value)
        return {self._chmap[socket]: Direction(e) for socket, e in self._poller.poll(timeout)}
    def warm_up(self) -> None:
        """Create and set up ZMQ sockets for all registered channels that does not have socket.
        """
        for chn in self.channels.values():
            if chn.socket is None:
                chn.set_socket(self.ctx.socket(chn.socket_type.value))
                self._chmap[chn.socket] = chn
    def shutdown(self, *, forced: bool=False) -> None:
        """Close all managed channels.

        Calls unbind/disconnect on active channels, and clears all sessions.

        Arguments:
            forced: When True, channels are closed with zero LINGER and all ZMQ errors are
                    ignored.
        """
        self._chmap = {}
        for chn in self.channels.values():
            if (self._poller is not None) and (chn.wait_for != Direction.NONE):
                self._poller.unregister(chn.socket)
            try:
                chn.on_shutdown(chn, forced)
            except:
                pass
            try:
                if chn.mode is SocketMode.BIND:
                    chn.unbind()
                elif chn.mode is SocketMode.CONNECT:
                    chn.disconnect()
            except:
                pass
            chn.sessions.clear()
            if forced:
                chn.drop_socket()
            else:
                chn.close_socket()


class Message(ABC):
    """Abstract base class for protocol message.
    """
    def __str__(self):
        return self.__class__.__qualname__
    __repr__ = __str__
    @abstractmethod
    def from_zmsg(self, zmsg: TZMQMessage) -> None:
        """Populate message data from sequence of ZMQ data frames.

        Arguments:
            zmsg: Sequence of frames that should be deserialized.

        Raises:
            InvalidMessageError: If message is not a valid protocol message.
        """
    @abstractmethod
    def as_zmsg(self) -> TZMQMessage:
        """Returns message as sequence of ZMQ data frames.
        """
    @abstractmethod
    def clear(self) -> None:
        """Clears message data.
        """
    @abstractmethod
    def copy(self) -> Message:
        """Returns copy of the message.
        """
    @abstractmethod
    def get_keys(self) -> Iterable:
        """Returns iterable of dictionary keys to be used with `Protocol.handlers`.
        Keys must be provided in order of precedence (from more specific to general).
        """

class SimpleMessage(Message):
    """Simple protocol message that holds items from ZMQ multipart message in its
    :attr:`.data` attribute.
    """
    def __init__(self):
        #: Sequence of data frames
        self.data: List[bytes] = []
    def from_zmsg(self, zmsg: TZMQMessage) -> None:
        """Populate message data from sequence of ZMQ data frames.

        Arguments:
            zmsg: Sequence of frames that should be deserialized.

        Raises:
            InvalidMessageError: If message is not a valid protocol message.

        Important:
            This class just makes a copy of items from ZMQ message list into :attr:`data`.
            All `~zmq.Frame` items are 'unpacked' into bytes, other items are simply copied.
        """
        self.data = [i.bytes if isinstance(i, zmq.Frame) else i for i in zmsg]
    def as_zmsg(self) -> TZMQMessage:
        """Returns message as sequence of ZMQ data frames.

        Important:
            This class simply returns the list kept in :attr:`.data` attribute. This may
            cause problems if returned list is subsequently updated. In such a case, create
            a copy of returned list, or create a subclass that overrides this method.
        """
        return self.data
    def clear(self) -> None:
        """Clears message data.
        """
        self.data.clear()
    def copy(self) -> Message:
        """Returns copy of the message.
        """
        msg = Message()
        msg.data = self.data.copy()
        return msg
    def get_keys(self) -> Iterable:
        """Returns iterable of dictionary keys to be used with `Protocol.handlers`.

        The default implementation returns list with first data frame or None, ANY sentinel.
        """
        return [self.data[0] if self.data else None, ANY]

class Session:
    """Base Peer Session class.
    """
    def __init__(self):
        #: Channel routing ID for connected peer
        self.routing_id: RoutingID = None
        #: Connected endpoint address, if any
        self.endpoint: Optional[ZMQAddress] = None
        #: Flag indicating that session is waiting for send
        self.send_pending: bool = False
    def __str__(self):
        return self.logging_id
    __repr__ = __str__
    @property
    def logging_id(self) -> str:
        "Returns _logging_id_ or <class_name>[<routing_id>:<endpoint>]"
        return getattr(self, '_logging_id_',
                       f'{self.__class__.__name__}[{self.routing_id}:{self.endpoint}]')

class Protocol(TracedMixin):
    """Base class for protocol.

    The main purpose of protocol class is to validate ZMQ messages, create protocol
    messages and session objects used by `Channel` to manage transmissions, and to handle
    messages received from channel. This base class defines common interface for message
    convertsion and validation. Descendant classes typically add methods for creation
    of protocol messages and message handling.
    """
    #: string with protocol OID (dot notation). MUST be set in child class.
    OID: str = '1.3.6.1.4.1.53446.1.5' # iso.org.dod.internet.private.enterprise.firebird.butler.protocol
    #: UUID instance that identifies the protocol. MUST be set in child class.
    UID: uuid.UUID = uuid.uuid5(uuid.NAMESPACE_OID, OID)
    #: Protocol revision (default 1)
    REVISION: int = 1
    def __init__(self, *, session_type: Type[Session]=Session):
        """
        Arguments:
            session_type: Class for session objects.
        """
        self._session_type: Type[Session] = session_type
        #: Callable that returns `.Message` instance
        self.message_factory = self.__message_factory
        #: Message handlers
        self.handlers: Dict[Any, TMessageHandler] = {}
    def __message_factory(self, zmsg: TZMQMessage=None) -> Message:
        "Internal message factory"
        return SimpleMessage()
    def __str__(self):
        return self.logging_id
    __repr__ = __str__
    def is_valid(self, zmsg: TZMQMessage) -> bool:
        """Return True if ZMQ message is a valid protocol message, otherwise returns False.

        Exceptions other than `~saturnin.core.types.InvalidMessageError` are not caught.

        Arguments:
            zmsg: ZeroMQ multipart message.
        """
        try:
            self.validate(zmsg)
        except InvalidMessageError:
            return False
        else:
            return True
    def validate(self, zmsg: TZMQMessage) -> None:
        """Verifies that sequence of ZMQ data frames is a valid protocol message.

        If this validation passes without exception, then `.convert_msg()` of the same
        message must be successful as well.

        Important:
            Implementation in base Protocol performs no validation and always returns True.

        Arguments:
            zmsg:   ZeroMQ multipart message.

        Raises:
            InvalidMessageError: If ZMQ message is not a valid protocol message.
        """
    def convert_msg(self, zmsg: TZMQMessage) -> Message:
        """Converts ZMQ message into protocol message.

        Arguments:
            zmsg: ZeroMQ multipart message.

        Returns:
            New protocol message instance with parsed ZMQ message. The base Protocol
            implementation returns `.Message` instance created by message factory.

        Raises:
            InvalidMessageError: If message is not a valid protocol message.
        """
        msg = self.message_factory(zmsg)
        msg.from_zmsg(zmsg)
        return msg
    def accept_new_session(self, channel: Channel, routing_id: RoutingID, msg: Message) -> bool:
        """Validates incoming message that initiated new session/transmission.

        Important:
            Default implementation unconditionally accept new sessions (always returns True).

        Arguments:
            channel:    Channel that received the message.
            routing_id: Routing ID of the sender.
            msg:        Received message.
        """
        return True
    def connect_with_session(self, channel: Channel) -> bool:
        """Called by :meth:`Channel.connect` to determine whether new session should be
        associated with connected peer.

        Note:
            Because it's not possible to call `Channel.send` without session, all protocols
            that require connecting peers to send a message to initiate transmission must
            return True.

            The default implementation uses :attr:`Channel.direction` to determine the
            return value (True if it contains `Direction.OUT`, else False).
        """
        return Direction.OUT in channel.direction
    def initialize_session(self, session: Session) -> None:
        """Initialize new session instance. The default implementation does nothing.
        """
    def handle_msg(self, channel: Channel, session: Session, msg: Message) -> Any:
        """Process message received from peer.

        Arguments:
            channel: Channel that received the message.
            session: Session instance.
            msg:     Received message.

        Uses :attr:`handlers` dictionary and `.Message.get_keys()` to select and execute
        the message handler. Exceptions raised by message handler are processed by
        `on_exception` event handler (if assigned). Exceptions raised by event handler are
        ignored, only `RuntimeWarning` is emitted.

        Returns:
            Whatever handler returns, or None when handler raises an exception.
        """
        handler: TMessageHandler = None
        for key in msg.get_keys():
            if handler := self.handlers.get(key):
                break
        try:
            assert handler, "Message without handler"
            return handler(channel, session, msg)
        except Exception as exc:
            try:
                self.handle_exception(channel, session, msg, exc)
            except:
                warnings.warn('Exception raised in exception handler', RuntimeWarning)
        return INVALID
    def handle_invalid_msg(self, channel: Channel, session: Session, exc: InvalidMessageError) -> None:
        """Called by `.Channel.receive()` when message conversion raises `InvalidMessageError`.

        Important:
            Executes `.on_invalid_msg` event. Descendant classes that override this method
            must call super() or execute this event directly.

            If this method is not overriden by descendant, and handler for this event is
            not defined, all `InvalidMessageError` exceptions are silently ignored.
        """
        self.on_invalid_msg(channel, session, exc)
    def handle_exception(self, channel: Channel, session: Session, msg: Message, exc: Exception) -> None:
        """Called by `.handle_msg()` on exception in message handler.

        Important:
            Executes `.on_exception` event. Descendant classes that override this method
            must call super() or execute this event directly.

            If this method is not overriden by descendant, and handler for this event is
            not defined, all exceptions raised in message handlers are silently ignored.
        """
        self.on_exception(channel, session, msg, exc)
    @eventsocket
    def message_factory(self, zmsg: TZMQMessage=None) -> Message:
        """Message factory that must return protocol message instance.
        The default factory produces new `SimpleMessage` instance on each call.
        """
    @eventsocket
    def on_invalid_msg(self, channel: Channel, session: Session, exc: InvalidMessageError) -> None:
        """Event called by `.Channel.receive()` when message conversion raises `InvalidMessageError`.

        Important:
            If handler for this event is not defined, all `InvalidMessageError` exceptions
            are silently ignored.
        """
    @eventsocket
    def on_exception(self, channel: Channel, session: Session, msg: Message, exc: Exception) -> None:
        """Event called by `.handle_msg()` on exception in message handler.

        Important:
            If handler for this event is not defined, all exceptions raised in message
            handlers are silently ignored.

            The exception thrown in this event handler is also not handled, and propagates
            to the upper layers (usually an I/O loop).
        """
    @property
    def session_type(self) -> Type[Session]:
        "Class for session objects."
        return self._session_type
    @property
    def logging_id(self) -> str:
        "Returns _logging_id_ or <class_name>[<uid>/<revision>]"
        return getattr(self, '_logging_id_', f'{self.__class__.__name__}[{self.UID.hex}/{self.REVISION}]')

class Channel(TracedMixin):
    """Base Class for ZMQ communication channel (socket).
    """
    def __init__(self, mngr: ChannelManager, name: str, protocol: Protocol,
                 routing_id: RoutingID, session_type: Type[Session], wait_for: Direction,
                 snd_timeout: int, rcv_timeout: int, linger: int, sock_opts: TSocketOptions):
        """
        Arguments:
            mngr: Channel manager.
            name: Channel name.
            routing_id: Routing ID for ZMQ socket.
            protocol: Protocol for serializing/deserializing messages.
            session_type: Session type. `DEFAULT` session type is obtained from Protocol.
            wait_for: Direction(s) of transmission events for this channel processed by
                      `ChannelManager.wait()`.
            snd_timeout: Timeout for send operation on the socket in milliseconds.
            rcv_timeout: Timeout for receive operation on the socket in milliseconds.
            linger: ZMQ socket linger period.
            sock_opts: Dictionary with socket options that should be set after socket creation.
        """
        self._mngr: ChannelManager = proxy(mngr)
        self._name = name
        self._routing_id: RoutingID = \
            uuid.uuid1().hex.encode() if routing_id is DEFAULT else routing_id
        self._protocol: Protocol = protocol
        self._session_type: Type[Session] = \
            protocol.session_type if session_type is DEFAULT else session_type
        self._snd_timeout: int = snd_timeout
        self._rcv_timeout: int = rcv_timeout
        self._linger: int = linger
        self._wait_for: Direction = wait_for
        self._mode: SocketMode = SocketMode.UNKNOWN
        self._socket_type: SocketType = SocketType.UNKNOWN_TYPE
        self._direction: Direction = Direction.BOTH
        #: ZMQ socket for transmission of messages
        self.socket: zmq.Socket = None
        #: Dictionary with socket options that should be set after socket creation
        self.sock_opts: TSocketOptions = sock_opts if sock_opts else dict()
        #: True if channel uses internal routing
        self.routed: bool = False
        #: List of binded/connected endpoints
        self.endpoints: List[ZMQAddress] = []
        #: Dictionary of active sessions, key=routing_id
        self.sessions: Dict[RoutingID, Session] = {}
        self._adjust()
    def __str__(self):
        return self.logging_id
    __repr__ = __str__
    def _adjust(self):
        """Called by `__init__()` to configure the channel parameters.
        """
    def set_socket(self, socket: zmq.Socket) -> None:
        """Used by `ChannelManager` to set socket to be used by `Channel`.
        """
        assert self.socket is None, "Channel socket is already set"
        self.socket = socket
        if self._routing_id:
            self.socket.routing_id = self._routing_id
        self.socket.immediate = 1
        self.socket.sndtimeo = self._snd_timeout
        self.socket.rcvtimeo = self._rcv_timeout
        self.socket.linger = self._linger
        if self.sock_opts:
            for name, value in self.sock_opts.items():
                setattr(self.socket, name, value)
        self._configure()
    def _configure(self) -> None:
        """Called by `.set_socket()` to configure the 0MQ socket.
        """
    def close_socket(self) -> None:
        """Close the ZMQ socket.

        Note:
            This will not change the linger value for socket, so underlying socket may not
            close if there are undelivered messages. The socket is actually closed only
            after all messages are delivered or discarded by reaching the socket’s LINGER
            timeout.
        """
        if self.socket and not self.socket.closed:
            self.socket.close()
        self.socket = None
    def drop_socket(self) -> None:
        """Unconditionally drops the ZMQ socket and all pending messages (forces LINGER=0).

        Note:
            All ZMQ errors raised by this operation are silently ignored.
        """
        try:
            if self.socket and not self.socket.closed:
                self.socket.close(0)
        except ZMQError:
            pass
        self.socket = None
    def create_session(self, routing_id: RoutingID) -> Session:
        """Returns newly created session.

        Arguments:
            routing_id: Routing ID for new session.

        Raises:
            ChannelError: When session for specified `routing_id` already exists.
        """
        if routing_id in self.sessions:
            raise ChannelError(f"Session for route {routing_id} already exists")
        session = self._session_type()
        session.routing_id = routing_id
        self.sessions[routing_id] = session
        self.protocol.initialize_session(session)
        return session
    def discard_session(self, session: Session) -> None:
        """Discard session object.

        If `.Session.endpoint` value is set, it also disconnects channel from this endpoint.

        Arguments:
            session: The Session to be discarded.
        """
        assert session.routing_id in self.sessions
        if session.endpoint:
            self.disconnect(session.endpoint)
        del self.sessions[session.routing_id]
    def bind(self, endpoint: ZMQAddress) -> ZMQAddress:
        """Bind the 0MQ socket to an address.

        Returns:
            The endpoint address. The returned address MAY differ from original address
            when wildcard specification is used.

        Raises:
            ChannelError: On attempt to a) bind another endpoint for PAIR socket, or
                b) bind to already binded endpoint.
        """
        assert self.socket is not None
        assert self.mode != SocketMode.CONNECT
        if (self.socket.socket_type == SocketType.PAIR) and self.endpoints:
            raise ChannelError("Cannot open multiple endpoints for PAIR socket")
        if endpoint in self.endpoints:
            raise ChannelError(f"Endpoint '{endpoint}' already openned")
        self.socket.bind(endpoint)
        endpoint = ZMQAddress(str(self.socket.last_endpoint, 'utf8'))
        self._mode = SocketMode.BIND
        self.endpoints.append(endpoint)
        return endpoint
    def unbind(self, endpoint: ZMQAddress=None) -> None:
        """Unbind from an address (undoes a call to `bind()`).

        Arguments:
            endpoint: Endpoint address, or None to unbind from all binded endpoints.
                      Note: The address must be the same as the addresss returned by `.bind()`.

        Raises:
            ChannelError: If channel is not binded to specified `endpoint`.
        """
        assert self.socket is not None
        assert self.mode == SocketMode.BIND
        if endpoint and endpoint not in self.endpoints:
            raise ChannelError(f"Endpoint '{endpoint}' not binded")
        addrs = [endpoint] if endpoint else list(self.endpoints)
        for addr in addrs:
            self.socket.unbind(addr)
            self.endpoints.remove(addr)
        if not self.endpoints:
            self._mode = SocketMode.UNKNOWN
    def connect(self, endpoint: ZMQAddress, *, routing_id: RoutingID=None) -> Optional[Session]:
        """Connect to a remote channel.

        Arguments:
            endpoint:   Endpoint address for connected peer.
            routing_id: Channel routing ID (required for routed channels).

        Returns:
            Session associated with connected peer, or None if no session was created.

        Raises:
            ChannelError: On attempt to a) connect another endpoint for PAIR socket, or
                b) connect to already connected endpoint.
        """
        assert self.socket is not None
        assert self.mode != SocketMode.BIND
        if (self.socket.socket_type == SocketType.PAIR) and self.endpoints:
            raise ChannelError("Cannot connect multiple endpoints for PAIR socket")
        if endpoint in self.endpoints:
            raise ChannelError(f"Endpoint '{endpoint}' already connected")
        if self.routed:
            assert routing_id
        else:
            routing_id = INTERNAL_ROUTE
        session = None
        if self.protocol.connect_with_session(self):
            session = self.create_session(routing_id)
            session.endpoint = endpoint
        if self.routed and routing_id:
            self.socket.connect_routing_id = routing_id
        self.socket.connect(endpoint)
        self._mode = SocketMode.CONNECT
        self.endpoints.append(endpoint)
        return session
    def disconnect(self, endpoint: ZMQAddress=None) -> None:
        """Disconnect from a remote socket (undoes a call to `connect()`).

        Important:
            Does not discards sessions that are bound to any disconnected endpoint.
            Use :meth:`discard_session` to disconnect & discard associated session.

        Arguments:
            endpoint: Endpoint address or None to disconnect from all connected endpoints.
                      Note: The address must be the same as the addresss returned by
                      :meth:`connect`.

        Raises:
            ChannelError: If channel is not connected to specified `endpoint`.
        """
        assert self.socket is not None
        assert self.mode == SocketMode.CONNECT
        if endpoint and endpoint not in self.endpoints:
            raise ChannelError(f"Endpoint '{endpoint}' not openned")
        addrs = [endpoint] if endpoint else list(self.endpoints)
        for addr in addrs:
            self.socket.disconnect(addr)
            self.endpoints.remove(addr)
        if not self.endpoints:
            self._mode = SocketMode.UNKNOWN
    def can_send(self, timeout: int=0) -> bool:
        """Returns True if underlying ZMQ socket is ready to accept at least one outgoing
        message without blocking (or dropping it).

        Important:
            It may return True for some sockets although subsequent `send()` may fail or
            block. Typicall situation is ROUTER socket that is attached to multiple peers.

        Arguments:
            timeout: Timeout in milliseconds passed to socket poll() call.
        """
        return self.socket.poll(timeout, POLLOUT) == POLLOUT
    def message_available(self, timeout: int=0) -> bool:
        """Returns True if underlying ZMQ socket is ready to receive at least one message
        without blocking (or error).

        Arguments:
            timeout: Timeout in milliseconds passed to socket poll() call.
        """
        return self.socket.poll(timeout, POLLIN) == POLLIN
    def send(self, msg: Message, session: Session) -> int:
        """Sends protocol message.

        Arguments:
            msg: Message to be sent.
            session: Session to which the message belongs.

        Returns:
            Zero for success, or ZMQ error code.
        """
        result = 0
        zmsg = msg.as_zmsg()
        if self.routed:
            zmsg.insert(0, session.routing_id)
        try:
            self.send_zmsg(zmsg)
        except Again as exc:
            if self.on_send_later.is_set() and self.on_send_later(self, session, msg):
                result = 0
            else:
                result = exc.errno
        except ZMQError as exc:
            if self.on_send_failed.is_set() and self.on_send_failed(self, session, msg, exc.errno):
                result = 0
            else:
                result = exc.errno
        return result
    def send_zmsg(self, zmsg: TZMQMessage) -> None:
        """Sends ZMQ multipart message.

        Important:
            Does not handle any ZMQError exception.
        """
        assert self.socket is not None
        assert Direction.OUT in self.direction, "Call to send() on RECEIVE-only channel"
        self.socket.send_multipart(zmsg)
    def receive(self, timeout: int=None) -> Any:
        """Receive and process protocol message with assigned protocol.

        If protocol raises `InvalidMessageError` on message conversion, it calls
        `Protocol.on_invalid_msg` event handler (if defined) before message is dropped.
        Exceptions raised by event handler are ignored, only `RuntimeWarning` is emitted.

        If there no session is found route, it first calls `Protocol.accept_new_session()`,
        and message is handled only when new session is accepted.

        Arguments:
             timeout: The timeout (in milliseconds) to wait for message.

        Returns:
            Whatever protocol message handler returns, sentinel TIMEOUT when timeout
            expires, or sentinel INVALID when:
            a) received message was not valid protocol message, or b) handler raises
            an exception, or c) there is no session associated with peer and new session
            was not accepted by protocol.
        """
        if timeout is not None:
            if self.socket.poll(timeout, POLLIN) == 0:
                return TIMEOUT
        try:
            zmsg = self.receive_zmsg()
        except Again as exc:
            if not (self.on_receive_later.is_set() and self.on_receive_later(self)):
                raise
            else:
                return INVALID
        except ZMQError as exc:
            if not (self.on_receive_failed.is_set() and self.on_receive_failed(self, exc.errno)):
                raise
            else:
                return INVALID
        routing_id: RoutingID = zmsg.pop(0) if self.routed else INTERNAL_ROUTE
        session = self.sessions.get(routing_id)
        #
        try:
            msg = self._protocol.convert_msg(zmsg)
        except InvalidMessageError as exc:
            try:
                self._protocol.handle_invalid_msg(self, session, exc)
            except:
                warnings.warn('Exception raised in invalid message handler', RuntimeWarning)
            return INVALID
        #
        if session is None:
            # This is the first message received for transmission with this peer
            if not self._protocol.accept_new_session(self, routing_id, msg):
                return INVALID
            session = self.create_session(routing_id)
        #
        return self._protocol.handle_msg(self, session, msg)
    def receive_zmsg(self) -> TZMQMessage:
        """Receive ZMQ multipart message.
        """
        assert self.socket is not None
        assert Direction.IN in self.direction, "Call to receive() on SEND-only channel"
        return self.socket.recv_multipart()
    def is_active(self) -> bool:
        """Returns True if channel is active (binded or connected).
        """
        return bool(self.endpoints)
    def set_wait_in(self, value: bool) -> None:
        "Set/clear `Direction.IN` in `.wait_for`"
        if value:
            self.wait_for |= Direction.IN
        else:
            self.wait_for = (self.wait_for | Direction.IN) ^ Direction.IN
    def set_wait_out(self, value: bool, session: Session=None) -> None:
        """Set/clear `Direction.OUT` in `.wait_for`.

        Arguments:
            value: New value for wait_for_out flag.
            session: Related session.

        Raises:
            ChannelError: For routed channel with active sessions when session is not provided.

        Important:
            If channel has active sessions, the `Session.send_pending` flag is also altered.
        """
        if session is None and self.sessions:
            if self.routed:
                raise ChannelError("Session required for routed channel")
            session = self.session
        if value:
            self.wait_for |= Direction.OUT
        else:
            self.wait_for = (self.wait_for | Direction.OUT) ^ Direction.OUT
        if session is not None:
            session.send_pending = value
    def wait(self, timeout: int=None) -> Direction:
        """Wait for socket events specified by :attr:`wait_for`.

        Arguments:
            timeout: The timeout (in milliseconds) to wait for an event. If unspecified,
                     will wait forever for an event.
        """
        assert self.socket is not None
        return Direction(self.socket.poll(timeout, self._wait_for.value))
    @property
    def name(self) -> str:
        "Channel name."
        return self._name
    @property
    def socket_type(self) -> SocketType:
        "ZMQ socket type this channel uses."
        return self._socket_type
    @property
    def direction(self) -> Direction:
        "Possible directions of transmission over this channel."
        return self._direction
    @property
    def mode(self) -> SocketMode:
        "ZMQ Socket mode."
        return self._mode
    @property
    def manager(self) -> ChannelManager:
        "The channel manager to which this channel belongs."
        return self._mngr
    @property
    def routing_id(self) -> RoutingID:
        """Routing_id value for ZMQ socket.
        """
        return self._routing_id
    @routing_id.setter
    def routing_id(self, value) -> None:
        if self.is_active():
            raise ChannelError("Cannot set routing_id of active channel")
        self._routing_id = value
        if self.socket is not None:
            self.socket.routing_id = value
    @property
    def protocol(self) -> Protocol:
        "Protocol used by channel"
        return self._protocol
    @property
    def session(self) -> Session:
        """Session associated with channel.

        Important: Valid *only* when channel has exactly one associated session.
        """
        assert len(self.sessions) == 1
        return self.sessions[INTERNAL_ROUTE]
    @property
    def snd_timeout(self) -> int:
        "Timeout for send operations."
        return self._snd_timeout
    @snd_timeout.setter
    def snd_timeout(self, value: int) -> None:
        self.socket.sndtimeo = value
        self._snd_timeout = value
    @property
    def rcv_timeout(self) -> int:
        "Timeout for receive operations."
        return self._rcv_timeout
    @rcv_timeout.setter
    def rcv_timeout(self, value: int) -> None:
        self.socket.rcvtimeo = value
        self._rcv_timeout = value
    @property
    def wait_for(self) -> Direction:
        """Direction(s) of transmission events for this channel processed by
        `ChannelManager.wait()` or `Channel.wait()`.

        Raises:
            ChannelError: When assigned value contains direction not supported by channel
                          for transmission.
        """
        return self._wait_for
    @wait_for.setter
    def wait_for(self, value: Direction) -> None:
        if not value in self.direction:
            raise ChannelError("Cannot wait for events in direction not supported "
                               "by channel for transmission.")
        self._wait_for = value
        self._mngr.update_poller(self, value)
    @property
    def logging_id(self) -> str:
        "Returns _logging_id_ or <class_name>[<name>]"
        return getattr(self, '_logging_id_', f'{self.__class__.__name__}[{self.name}]')
    @property
    def log_context(self) -> Any:
        "Logging context. Returns `log_context` from ChannelManager."
        return self._mngr.log_context
    @eventsocket
    def on_output_ready(self, channel: Channel) -> None:
        """Event called when channel is ready to accept at least one outgoing message
        without blocking (or dropping it).

        Arguments:
            channel: Channel ready for sending a message.
        """
    @eventsocket
    def on_shutdown(self, channel: Channel, forced: bool) -> None:
        """Event called by :meth:`ChannelManager.shutdown` before the channel is shut down.

        Important:
            All exceptions escaping this method are silently ignored.

        Arguments:
            channel: Channel to be shut down.
            forced:  When True, the channel will be closed with zero LINGER and all ZMQ
                     errors will be ignored.
        """
    @eventsocket
    def on_send_failed(self, channel: Channel, session: Session, msg: Message, err_code: int) -> bool:
        """Event called by :meth:`Channel.send` when send operation fails with `zmq.ZMQError`
        exception other than EAGAIN.

        If event returns True, the error is ignored, otherwise the error code is reported
        to the caller.
        """
    @eventsocket
    def on_send_later(self, channel: Channel, session: Session, msg: Message) -> bool:
        """Event called by :meth:`Channel.send` when send operation fails with zmq.Again`
        exception.

        If event returns True, the error is ignored, otherwise the error code is reported
        to the caller.
        """
    @eventsocket
    def on_receive_failed(self, channel: Channel, err_code: int) -> bool:
        """Event called by :meth:`Channel.receive` when receive operation fails with
        `zmq.ZMQError` exception other than EAGAIN.

        If event returns True, the receive() returns INVALID, otherwise the exception is
        propagated to the caller.
        """
    @eventsocket
    def on_receive_later(self, channel: Channel) -> bool:
        """Event called by :meth:`Channel.receive` when receive operation fails with
        `zmq.Again` exception.

        If event returns True, the receive() returns INVALID, otherwise the exception is
        propagated to the caller.
        """

# Channels for individual ZMQ socket types
class DealerChannel(Channel):
    """Communication channel over DEALER socket.
    """
    def _adjust(self):
        self._socket_type = SocketType.DEALER

class PushChannel(Channel):
    """Communication channel over PUSH socket.
    """
    def _adjust(self):
        self._socket_type = SocketType.PUSH
        self._direction = Direction.OUT

class PullChannel(Channel):
    """Communication channel over PULL socket.
    """
    def _adjust(self):
        self._socket_type = SocketType.PULL
        self._direction = Direction.IN

class PubChannel(Channel):
    """Communication channel over PUB socket.
    """
    def _adjust(self):
        self._socket_type = SocketType.PUB
        self._direction = Direction.OUT

class SubChannel(Channel):
    """Communication channel over SUB socket.
    """
    def _adjust(self):
        self._socket_type = SocketType.SUB
        self._direction = Direction.IN
    def subscribe(self, topic: bytes):
        """Subscribe to topic.

        Arguments:
            topic: ZMQ topic.
        """
        self.socket.subscribe = topic
    def unsubscribe(self, topic: bytes):
        """Unsubscribe from topic.

        Arguments:
            topic: ZMQ topic.
        """
        self.socket.unsubscribe = topic

class XPubChannel(Channel):
    """Communication channel over XPUB socket.
    """
    def _adjust(self):
        self._socket_type = SocketType.XPUB
    def _configure(self):
        "Create XPUB socket for this channel."
        super()._configure()
        self.socket.xpub_verboser = 1 # pass subscribe and unsubscribe messages on XPUB socket

class XSubChannel(Channel):
    """Communication channel over XSUB socket.
    """
    def _adjust(self):
        self._socket_type = SocketType.XSUB
        self._direction = Direction.IN
    def subscribe(self, topic: bytes):
        """Subscribe to topic.

        Arguments:
            topic: ZMQ topic.
        """
        self.socket.send_multipart(b'\x01', topic)
    def unsubscribe(self, topic: bytes):
        """Unsubscribe to topic.

        Arguments:
            topic: ZMQ topic.
        """
        self.socket.send_multipart(b'\x00', topic)

class PairChannel(Channel):
    """Communication channel over PAIR socket.
    """
    def _adjust(self):
        self._socket_type = SocketType.PAIR

class RouterChannel(Channel):
    """Communication channel over ROUTER socket.
    """
    def _adjust(self):
        self._socket_type = SocketType.ROUTER
        self.routed = True
    def _configure(self):
        super()._configure()
        self.socket.router_mandatory = 1


# Types
TZMQMessage = List[Union[bytes, Frame]]
"ZMQ multipart message"
TMessageFactory = Callable[[Optional[TZMQMessage]], Message]
"Message factory callable"
TSocketOptions = Dict[str, Any]
"ZMQ socket options"
TMessageHandler = Callable[[Channel, Session, Message], None]
"Message handler"



#coding:utf-8
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/lib/data/provider.py
# DESCRIPTION:    Base class for Saturnin data provider microservices
# CREATED:        14.12.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2020 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Saturnin base class for data provider microservices
"""

from __future__ import annotations
from typing import cast
from functools import partial
from firebird.base.config import MIME, StrOption, EnumOption, \
     IntOption, BoolOption, ZMQAddressOption, MIMEOption
from saturnin.base import Error, StopError, Direction, SocketMode, PipeSocket, Outcome, \
     ZMQAddress, MIME, ComponentConfig, Channel, DealerChannel
from saturnin.component.micro import MicroService
from saturnin.protocol.fbdp import ErrorCode, FBDPServer, FBDPClient, \
     FBDPSession, FBDPMessage

PIPE_CHN = 'pipe'

class DataProviderConfig(ComponentConfig):
    """Base data provider microservice configuration."""
    def __init__(self, name: str):
        super().__init__(name)
        self.stop_on_close: BoolOption = \
            BoolOption('stop_on_close', "Stop service when pipe is closed", default=True)
        self.pipe: StrOption = \
            StrOption('pipe', "Data Pipe Identification", required=True)
        self.pipe_address: ZMQAddressOption = \
            ZMQAddressOption('pipe_address', "Data Pipe endpoint address", required=True)
        self.pipe_mode: EnumOption = \
            EnumOption('pipe_mode', SocketMode, "Data Pipe Mode", required=True)
        self.pipe_format: MIMEOption = \
            MIMEOption('pipe_format', "Pipe data format specification")
        self.batch_size: IntOption = \
            IntOption('batch_size', "Data batch size", required=True, default=50)
        self.ready_schedule_interval: IntOption = \
            IntOption('ready_schedule_interval',
                      "READY message schedule interval in milliseconds", required=True,
                      default=1000)
    def validate(self) -> None:
        """Extended validation.

        - `pipe_format` is required for CONNECT `pipe_mode`.
        """
        super().validate()
        if self.pipe_mode.value is SocketMode.CONNECT and self.pipe_format.value is None:
            raise Error("'pipe_format' required for CONNECT pipe mode.")

class DataProviderMicro(MicroService):
    """Base data provider microservice.

    Descendant classes should override:

    - `handle_accept_client` to validate client request and aquire resources associated
      with pipe.
    - `handle_produce_data` to produce data for outgoing DATA message.
    - `handle_pipe_closed` to release resource assiciated with pipe.
    """
    def initialize(self, config: DataProviderConfig) -> None:
        """Verify configuration and assemble component structural parts.
        """
        super().initialize(config)
        # PipeSocket.OUTPUT
        # Configuration
        self.stop_on_close = config.stop_on_close.value
        self.pipe: str = config.pipe.value
        self.pipe_mode: SocketMode = config.pipe_mode.value
        self.pipe_address: ZMQAddress = config.pipe_address.value
        self.pipe_format: MIME = config.pipe_format.value
        self.batch_size: int = config.batch_size.value
        self.ready_schedule_interval: int = config.ready_schedule_interval.value
        #
        if self.pipe_mode == SocketMode.BIND:
            self.protocol = FBDPServer()
            self.protocol.on_exception = self.handle_exception
            self.protocol.on_accept_client = self.handle_accept_client
            self.protocol.on_schedule_ready = self.handle_schedule_ready
            # We have an endpoint to bind
            self.endpoints[PIPE_CHN] = [self.pipe_address]
        else:
            self.protocol = FBDPClient()
        self.protocol.batch_size = self.batch_size
        self.protocol.on_pipe_closed = self.handle_pipe_closed
        self.protocol.on_produce_data = self.handle_produce_data
        # High water mark optimization
        rcvhwm = 5
        sndhwm = self.batch_size + 5
        # Create pipe channel
        chn: Channel = self.mngr.create_channel(DealerChannel, PIPE_CHN, self.protocol,
                                                wait_for=Direction.IN,
                                                sock_opts={'rcvhwm': rcvhwm,
                                                           'sndhwm': sndhwm,})
        chn.protocol.log_context = self.logging_id
    def aquire_resources(self) -> None:
        """Aquire resources required by component (open files, connect to other services etc.).

        Must raise an exception when resource aquisition fails.
        """
        # Connect to the data pipe
        if self.pipe_mode == SocketMode.CONNECT:
            chn: Channel = self.mngr.channels[PIPE_CHN]
            session = chn.connect(self.pipe_address)
            # OPEN the data pipe connection, this also fills session attributes
            # We are PRODUCER client, we must attach to server INPUT
            cast(FBDPClient, chn.protocol).send_open(chn, session, self.pipe,
                                                     PipeSocket.INPUT, self.pipe_format)
    def release_resources(self) -> None:
        """Release resources aquired by component (close files, disconnect from other services etc.)
        """
        # CLOSE all active data pipe sessions
        chn: Channel = self.mngr.channels[PIPE_CHN]
        # send_close() will discard session, so we can't iterate over sessions.values() directly
        for session in list(chn.sessions.values()):
            # We have to report error here, because normal is to close pipes before
            # shutdown is commenced. Mind that service shutdown could be also caused by error!
            cast(FBDPServer, chn.protocol).send_close(chn, session, ErrorCode.ERROR)
    def handle_exception(self, channel: Channel, session: FBDPSession, msg: FBDPMessage,
                         exc: Exception) -> None:
        """Event handler called by `.handle_msg()` on exception in message handler.

        Sets service outcome to ERROR and notes exception as details.
        """
        self.outcome = Outcome.ERROR
        self.details = exc
    # FBDP server only
    def handle_accept_client(self, session: FBDPSession) -> None:
        """Event handler executed when client connects to the data pipe via OPEN message.

        Arguments:
            session: Session associated with client.

        The session attributes `data_pipe`, `pipe_socket`, `data_format` and `params`
        contain information sent by client, and the event handler validates the request.

        If request should be rejected, it raises the `StopError` exception with `code`
        attribute containing the `ErrorCode` to be returned in CLOSE message.

        Important:
            Base implementation validates pipe identification and pipe socket, and converts
            data format from string to MIME (in session).

            The descendant class that overrides this method must call super() as first
            action.
        """
        if session.pipe != self.pipe:
            raise StopError(f"Unknown data pipe '{session.data_pipe}'",
                            code = ErrorCode.PIPE_ENDPOINT_UNAVAILABLE)
        # We're PRODUCER server, so clients can only attach to our OUTPUT
        elif session.socket is not PipeSocket.OUTPUT:
            raise StopError(f"'{session.socket}' socket not available",
                            code = ErrorCode.PIPE_ENDPOINT_UNAVAILABLE)
        # We work with MIME formats, so we'll convert the format specification to MIME
        session.data_format = MIME(session.data_format)
    def handle_schedule_ready(self, channel: Channel, session: FBDPSession, msg: FBDPMessage) -> None:
        """Event handler executed in order to send the READY message to the client later.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with client.
            msg:     Message that triggered the scheduling.

        The event handler may cancel the transmission by raising the `StopError` exception
        with `code` attribute containing the `ErrorCode` to be returned in CLOSE message.

        Important:
            The base implementation schedules `resend_ready()` according to
            `ready_schedule_interval` configuration option.
        """
        self.schedule(partial(cast(FBDPServer, channel.protocol).resend_ready,
                              channel, session, msg),
                      self.ready_schedule_interval)
    # FBDP common
    def handle_produce_data(self, channel: Channel, session: FBDPSession, msg: FBDPMessage) -> None:
        """Event handler executed to store data into outgoing DATA message.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with client.
            msg: DATA message that will be sent to client.

        The event handler must store the data in `msg.data_frame` attribute. It may also
        set ACK-REQUEST flag and `type_data` attribute.

        The event handler may cancel the transmission by raising the `StopError` exception
        with `code` attribute containing the `ErrorCode` to be returned in CLOSE message.

        Note:
            To indicate end of data, raise StopError with ErrorCode.OK code.

        Note:
            Exceptions are handled by protocol, but only StopError is checked for protocol
            ErrorCode. As we want to report INVALID_DATA properly, we have to convert
            UnicodeError into StopError.

        Important:
            The base implementation simply raises StopError with ErrorCode.OK code, so
            the descendant class must override this method without super() call.
        """
        raise StopError('OK', code=ErrorCode.OK)
    def handle_pipe_closed(self, channel: Channel, session: FBDPSession, msg: FBDPMessage,
                           exc: Exception=None) -> None:
        """Event handler executed when CLOSE message is received or sent, to release any
        resources associated with current transmission.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with peer.
            msg: Received/sent CLOSE message.
            exc: Exception that caused the error.

        Important:
            The base implementation does only two actions:

            - If exception is provided, sets service execution outcome to ERROR
              and notes exception in details.
            - Stops the service if `stop_on_close` is True.

            The descendant class that overrides this method must call super().
        """
        # FDBP converts exceptions raised in our event handler to CLOSE messages, so
        # here is the central place to handle errors in data pip processing.
        # Note problem in service execution outcome
        if exc is not None:
            self.outcome = Outcome.ERROR
            self.details = exc
        #
        if self.stop_on_close:
            self.stop.set()

